from collections import deque
from copy import copy
import sys
from types import NotImplementedType
from typing import cast, Generic, Iterable, SupportsIndex, TypeVar

_ObjT = TypeVar('_ObjT', bound=object)


class SetDeque(deque, Generic[_ObjT]):
    def __init__(self, iterable: Iterable = (), maxlen: int | None = None):
        unique_ordered_els = dict.fromkeys(iterable).keys()
        super().__init__(iterable=unique_ordered_els, maxlen=maxlen)
        self._set: set[_ObjT] = set(self.__iter__())

    def _add_element(self, method_name: str, *args: object, el: _ObjT) -> bool:
        if el not in self._set:
            getattr(super(), method_name)(*args, el)
            self._set.add(el)
            return True
        return False

    def _add_elements(self, method_name: str, el_iter: Iterable[_ObjT]) -> None:
        unique_ordered_els = dict.fromkeys(el_iter).keys()
        new_unique_ordered_els = (el for el in unique_ordered_els if el not in self._set)
        getattr(super(), method_name)(new_unique_ordered_els)
        self._set |= unique_ordered_els

    def _remove_returned_element(self, method_name: str) -> _ObjT:
        el = getattr(super(), method_name)()
        self._set.remove(el)
        return el

    def _remove_explicit_element(self,
                                 method_name: str,
                                 *args: object,
                                 el: _ObjT,
                                 add_el_to_args: bool) -> None:
        if add_el_to_args:
            args = args + (el,)
        getattr(super(), method_name)(*args)
        self._set.remove(el)

    def append(self, el: _ObjT) -> None:
        self._add_element('append', el=el)

    def appendleft(self, el: _ObjT) -> None:
        self._add_element('appendleft', el=el)

    def extend(self, el_iter: Iterable[_ObjT]) -> None:
        self._add_elements('extend', el_iter)

    def extendleft(self, el_iter: Iterable[_ObjT]) -> None:
        self._add_elements('extendleft', el_iter)

    def insert(self, __i: int, __x: _ObjT) -> None:
        self._add_element('insert', __i, el=__x)

    def pop(self) -> _ObjT:  # type: ignore[override]
        return self._remove_returned_element('pop')

    def popleft(self) -> _ObjT:
        return self._remove_returned_element('popleft')

    def remove(self, __value) -> None:
        self._remove_explicit_element('remove', el=__value, add_el_to_args=True)

    def clear(self) -> None:
        super().clear()
        self._set.clear()

    def count(self, __x: _ObjT) -> int:
        return 1 if __x in self._set else 0

    def __setitem__(self, __key: SupportsIndex, __value: _ObjT) -> None:  # type: ignore[override]
        old_val = super().__getitem__(__key)
        if __value != old_val:
            added = self._add_element('__setitem__', __key, el=__value)
            if added:
                self._set.remove(old_val)

    def __delitem__(self, __key: SupportsIndex) -> None:  # type: ignore[override]
        self._remove_explicit_element(
            '__delitem__',
            __key,
            el=super().__getitem__(__key),
            add_el_to_args=False,
        )

    def __add__(self, __value: 'SetDeque[_ObjT]') -> 'SetDeque[_ObjT] | NotImplementedType':
        if not isinstance(__value, SetDeque):
            return NotImplemented  # type: ignore[unreachable]
        self_copy = copy(self)
        self_copy.extend(__value)
        return self_copy

    def __iadd__(self, __value: Iterable[_ObjT]):  # type: ignore[misc]
        self.extend(__value)
        return self

    def __eq__(self, __value: object):
        return super().__eq__(__value) and self._set == cast(SetDeque[_ObjT], __value)._set

    def __reduce__(self):
        ret = super().__reduce__()
        return ret[0], ret[1], None, ret[3]

    def __contains__(self, __key: object) -> bool:
        return self._set.__contains__(__key)

    def _common_mul(self, __value: int):
        if __value > 0:
            return SetDeque[_ObjT](self.__iter__())
        else:
            return SetDeque[_ObjT]()

    def __mul__(self, __value: int):
        return self._common_mul(__value)

    def __rmul__(self, __value: int):
        return self._common_mul(__value)

    def __imul__(self, __value: int):
        if __value <= 0:
            self.clear()
        return self

    def __repr__(self):
        from typing import get_args
        arg_repr = ''
        if hasattr(self, '__orig_class__'):
            args = get_args(self.__orig_class__)
            if len(args) == 1:
                arg = args[0]
                arg_repr = f"[{arg.__name__ if hasattr(arg, '__name__') else repr(arg)}]"
        maxlen_repr = f', maxlen={self.maxlen}' if self.maxlen is not None else ''
        return f'{self.__class__.__name__}{arg_repr}({list(self)}{maxlen_repr})'

    def __sizeof__(self):
        return super().__sizeof__() + sys.getsizeof(self._set)



from typing import Protocol


class IsUniquelyNamedJob(Protocol):
    """"""
    @property
    def name(self) -> str:
        ...

    @property
    def unique_name(self) -> str:
        ...

    def __init__(self, *, name: str | None = None):
        ...

    def regenerate_unique_name(self) -> None:
        ...


class IsNestedContext(Protocol):
    """"""
    def __enter__(self):
        ...

    def __exit__(self, exc_type, exc_value, traceback):
        ...



from collections.abc import Callable, Iterable, Mapping, Sequence
from contextlib import contextmanager
from copy import copy, deepcopy
import functools
import inspect
import json
from types import GenericAlias, NoneType, UnionType
from typing import (Annotated,
                    Any,
                    cast,
                    ContextManager,
                    ForwardRef,
                    Generator,
                    Generic,
                    get_args,
                    get_origin,
                    Hashable,
                    Iterator,
                    Literal,
                    Optional,
                    SupportsIndex,
                    Union)

from pydantic import Protocol as PydanticProtocol
from pydantic import root_validator, ValidationError
from pydantic.error_wrappers import ErrorWrapper
from pydantic.fields import Field, ModelField, Undefined, UndefinedType
from pydantic.generics import GenericModel
from pydantic.main import BaseModel, ModelMetaclass, validate_model
from pydantic.typing import is_none_type
from pydantic.utils import lenient_isinstance, lenient_issubclass, sequence_like
from typing_extensions import get_original_bases, TypeVar

from omnipy.api.protocols.private.util import IsSnapshotWrapper
from omnipy.api.typedefs import TypeForm
from omnipy.data.data_class_creator import DataClassBase, DataClassBaseMeta
from omnipy.data.helpers import (cleanup_name_qualname_and_module,
                                 get_special_methods_info_dict,
                                 is_model_instance,
                                 MethodInfo,
                                 ResetSolutionTuple,
                                 validate_cls_counts,
                                 YesNoMaybe)
from omnipy.data.missing import parse_none_according_to_model
from omnipy.data.mixins.display import ModelDisplayMixin
from omnipy.util.contexts import (hold_and_reset_prev_attrib_value,
                                  LastErrorHolder,
                                  nothing,
                                  setup_and_teardown_callback_context)
from omnipy.util.decorators import add_callback_after_call, no_context
from omnipy.util.helpers import (all_equals,
                                 all_type_variants,
                                 ensure_plain_type,
                                 evaluate_any_forward_refs_if_possible,
                                 get_calling_module_name,
                                 get_default_if_typevar,
                                 is_non_omnipy_pydantic_model,
                                 is_optional,
                                 is_union,
                                 remove_forward_ref_notation)
from omnipy.util.setdeque import SetDeque

_KeyT = TypeVar('_KeyT', bound=Hashable)
_ValT = TypeVar('_ValT')
_IterT = TypeVar('_IterT')
_ReturnT = TypeVar('_ReturnT')
_IdxT = TypeVar('_IdxT', bound=SupportsIndex)
_RootT = TypeVar('_RootT')
_ModelT = TypeVar('_ModelT')

ROOT_KEY = '__root__'

# TODO: Refactor Dataset and Model using mixins (including below functions)


class _ModelMetaclass(DataClassBaseMeta, ModelMetaclass):
    # Hack to overcome bug in pydantic/fields.py (v1.10.13), lines 636-641:
    #
    # if origin is None or origin is CollectionsHashable:
    #     # field is not "typing" object eg. Union, dict, list etc.
    #     # allow None for virtual superclasses of NoneType, e.g. Hashable
    #     if isinstance(self.type_, type) and isinstance(None, self.type_):
    #         self.allow_none = True
    #     return
    #
    # This hinders models (including pure pydantic BaseModels) to be properly considered as
    # subfields, e.g. in `list[MyModel]` as `get_origin(MyModel) is None`. Here, we want allow_none
    # to be set to True so that Model is allowed to validate a None value.
    #
    # TODO: Revisit the need for _ModelMetaclass hack in pydantic v2
    def __instancecheck__(self, instance: Any) -> bool:
        if instance is None:
            return True
        return super().__instancecheck__(instance)


class Model(
        ModelDisplayMixin,
        DataClassBase,
        GenericModel,
        Generic[_RootT],
        metaclass=_ModelMetaclass,
):
    """
    A data model containing a value parsed according to the model.

    If no value is provided, the value is set to the default value of the data model, found by
    calling the model class without parameters, e.g. `int()`.

    Model is a generic class that cannot be instantiated directly. Instead, a Model class needs
    to be specialized with a data type before Model objects can be instantiated. A data model
    functions as a data parser and guarantees that the parsed data follows the specified model.

    Example data model specialized as a class alias::

        MyNumberList = Model[list[int]]

    ... alternatively as a Model subclass::

        class MyNumberList(Model[list[int]]):
            pass

    Once instantiated, a Model object functions as a parser, e.g.::

        my_number_list = MyNumberList([2,3,4])

        my_number_list.contents = ['3', 4, True]
        assert my_number_list.contents == [3,4,1]

    While the following should raise a `ValidationError`::

        my_number_list.contents = ['abc', 'def']

    The Model class is a wrapper class around the powerful `GenericModel` class from pydantic.

    See also docs of the Dataset class for more usage examples.
    """

    __root__: _RootT = Field(default_factory=lambda: Undefined)

    class Config:
        arbitrary_types_allowed = True
        validate_all = True
        # validate_assignment = True
        smart_union = True
        # json_loads = orjson.loads
        # json_dumps = orjson_dumps

    @classmethod
    def _get_default_factory_from_model(
            cls, model: type[_RootT] | TypeForm | TypeVar) -> Callable[[], _RootT]:
        default_val = cls._get_default_value_from_model(model)

        def default_factory() -> _RootT:
            return default_val

        return default_factory

    @classmethod
    def _get_default_value_from_model(  # noqa: C901
            cls,
            model: type[_RootT] | TypeForm | TypeVar,
    ) -> _RootT:
        model = get_default_if_typevar(model)
        origin_type = get_origin(model)
        args = get_args(model)

        if origin_type is Annotated:
            model = args[0]
            origin_type = get_origin(model)
            args = get_args(model)

        if origin_type in (None, ()):
            origin_type = model

        if origin_type is None:
            origin_type = NoneType

        if origin_type in [Union, UnionType]:
            if any(is_none_type(arg) for arg in args):
                return cast(_RootT, None)

            last_error_holder = LastErrorHolder()
            for arg in args:
                if callable(arg) or isinstance(arg, TypeVar):
                    with last_error_holder:
                        return cls._get_default_value_from_model(arg)
            last_error_holder.raise_derived(TypeError(f'Cannot instantiate model "{model}".'))

        if origin_type is tuple and args and Ellipsis not in args:
            return cast(_RootT, tuple(cls._get_default_value_from_model(arg) for arg in args))

        if origin_type is Literal:
            return args[0]

        if origin_type is Callable:
            return cast(_RootT, lambda: None)

        if origin_type is ForwardRef or type(origin_type) is ForwardRef:
            raise TypeError(f'Cannot instantiate model "{model}". ')

        return cast(_RootT, origin_type())

    @classmethod
    def _prepare_cls_members_to_mimic_model(cls, created_model: 'Model[type[_RootT]]') -> None:
        outer_type = created_model.outer_type(with_args=True)
        outer_type_plain = created_model.outer_type(with_args=False)

        # TODO: See if it is possible to type Model mimicking of root type, e.g. with Protocol
        if not lenient_issubclass(outer_type_plain, TypeVar):
            if is_union(outer_type) or outer_type_plain is Literal:
                outer_types = get_args(outer_type)
            else:
                outer_types = (outer_type_plain,)

            for name, method_info in get_special_methods_info_dict().items():
                names_to_check = (name, '__add__') if name in ('__iadd__', '__radd__') else (name,)
                for type_to_support in outer_types:
                    for name_to_check in names_to_check:
                        setattr(created_model,
                                name,
                                functools.partialmethod(cls._special_method, name, method_info))
                        break
                    else:
                        continue
                    # To let the inner break, also break the outer for loop
                    break

    def __class_getitem__(  # type: ignore[override]
        cls,
        params: type[_RootT] | tuple[type[_RootT]] | tuple[type[_RootT], Any] | TypeVar
        | tuple[TypeVar, ...],
    ) -> 'Model[type[_RootT]]':

        model = cls._prepare_params(params)

        orig_model: type[_RootT] | tuple[type[_RootT], Any] | TypeVar = model

        # Populating the root field at runtime instead of providing a __root__ Field explicitly
        # is needed due to the inability of typing/pydantic to provide a dynamic default based on
        # the actual type. The following issue in mypy seems relevant:
        # https://github.com/python/mypy/issues/3737 (as well as linked issues)

        created_model = cast(Model, super().__class_getitem__(model if cls == Model else params))

        created_model._get_root_field().field_info = deepcopy(
            created_model._get_root_field().field_info)

        if cls is Model and orig_model is not _RootT:
            created_model._get_root_field().field_info.extra = {'orig_model': orig_model}

        created_model._inherit_first_orig_model_in_bases_if_missing()

        cls._recursively_set_allow_none(created_model._get_root_field())

        # As long as models are not created concurrently, setting the class members temporarily
        # should not have averse effects
        # TODO: Check if we can move to explicit definition of __root__ field at the object
        #       level in pydantic 2.0 (when it is released)

        if created_model is not cls:
            cleanup_name_qualname_and_module(cls, created_model, orig_model)

        if not isinstance(model, TypeVar):
            cls._prepare_cls_members_to_mimic_model(created_model)

        return created_model

    @classmethod
    def _inherit_first_orig_model_in_bases_if_missing(cls):
        if cls is not Model:
            for orig_base in get_original_bases(cls):
                if isinstance(orig_base, _ModelMetaclass) and orig_base.__concrete__:
                    orig_base._inherit_first_orig_model_in_bases_if_missing()
                    orig_model = orig_base.get_orig_model()
                    if orig_model is not Undefined:
                        cls.set_orig_model(orig_model)
                        break

            cls._clean_type_caches()

    def __new__(
        cls,
        *args: Any,
        **kwargs: Any,
    ) -> 'Model[_RootT]':
        model_not_specified = ROOT_KEY not in cls.__fields__
        if model_not_specified:
            cls._raise_no_model_exception()

        return super().__new__(cls)

    @classmethod
    def get_orig_model(cls) -> type[_RootT] | UndefinedType:
        if cls.__fields__[ROOT_KEY].field_info and cls.__fields__[ROOT_KEY].field_info.extra:
            return cls.__fields__[ROOT_KEY].field_info.extra.get('orig_model', Undefined)
        return Undefined

    @classmethod
    def set_orig_model(cls, orig_model: type[_RootT]) -> None:
        cls.__fields__[ROOT_KEY].field_info.extra['orig_model'] = orig_model

    def __init__(  # noqa: C901
        self,
        value: Any | UndefinedType = Undefined,
        *,
        __root__: Any | UndefinedType = Undefined,
        **kwargs: Any,
    ) -> None:
        super_kwargs: dict[str, _RootT] = {}
        num_root_vals = 0

        if value is not Undefined:
            super_kwargs[ROOT_KEY] = cast(_RootT, value)
            num_root_vals += 1

        if __root__ is not Undefined:
            super_kwargs[ROOT_KEY] = cast(_RootT, __root__)
            num_root_vals += 1

        if kwargs:
            super_kwargs[ROOT_KEY] = cast(_RootT, kwargs)
            kwargs = {}
            num_root_vals += 1

        assert num_root_vals <= 1, 'Not allowed to provide root data in more than one argument'

        if self._get_root_field().default_factory() is Undefined:
            default_factory = self.__class__._get_default_factory_from_model(self.full_type())
            self._get_root_field().default_factory = default_factory

        omnipy_or_pydantic_model_as_input = False
        if ROOT_KEY in super_kwargs:
            try:
                omnipy_or_pydantic_model_as_input, value = \
                    self._prepare_value_for_validation_if_model(super_kwargs[ROOT_KEY])
            except Exception as exc:
                val_exc = ValueError(f'Failed to prepare value for validation: {exc}')
                raise ValidationError(
                    [ErrorWrapper(exc, loc=ROOT_KEY), ErrorWrapper(val_exc, loc=ROOT_KEY)],
                    self.__class__)
            if omnipy_or_pydantic_model_as_input:
                super_kwargs[ROOT_KEY] = cast(_RootT, value)

        self._init(super_kwargs, **kwargs)

        try:
            self._primary_validation(super_kwargs)
        except ValidationError:
            if omnipy_or_pydantic_model_as_input:
                self._secondary_validation_from_data(super_kwargs)
            else:
                raise

        if not self.__class__.__doc__:
            self._set_standard_field_description()

    def _primary_validation(self, super_kwargs):
        # Pydantic validation of super_kwargs
        validate_cls_counts[self.__class__.__name__] += 1
        super().__init__(**super_kwargs)

    def _secondary_validation_from_data(self, super_kwargs):
        super().__init__()
        self.from_data(super_kwargs[ROOT_KEY])

    def _prepare_value_for_validation_if_model(self, value: object) -> tuple[bool, object]:
        if is_model_instance(value):
            return True, cast(Model[_RootT], value).to_data()
        elif is_non_omnipy_pydantic_model(value):
            return True, cast(BaseModel, value).dict(by_alias=True)
        return False, value

    def _init(self, super_kwargs: dict[str, Any], **kwargs: Any) -> None:
        ...

    def __del__(self):
        contents_id = id(self.contents)
        self.snapshot_holder.schedule_deepcopy_content_ids_for_deletion(contents_id)

    def __copy__(self):
        return self.copy(deep=False)

    def copy(self, *, deep: bool = False, **kwargs) -> 'Model[_RootT]':
        pydantic_copy = GenericModel.copy(self, deep=deep, **kwargs)
        if not deep:
            pydantic_copy.__dict__[ROOT_KEY] = pydantic_copy.__dict__[ROOT_KEY].copy()
        return pydantic_copy

    @classmethod
    def clone_model_cls(cls: type[_ModelT], new_model_cls_name: str) -> type[_ModelT]:
        new_model_cls: type[_ModelT] = type(new_model_cls_name, (cls,), {})
        return new_model_cls

    @staticmethod
    def _raise_no_model_exception() -> None:
        raise TypeError('Note: The Model class requires a concrete model to be specified as '
                        'a type hierarchy within brackets either directly, e.g.:\n\n'
                        '\tmodel = Model[list[int]]([1,2,3])\n\n'
                        'or indirectly in a subclass definition, e.g.:\n\n'
                        '\tclass MyNumberList(Model[list[int]]): ...\n\n')

    def _set_standard_field_description(self) -> None:
        self.__fields__[ROOT_KEY].field_info.description = self._get_standard_field_description()

    @classmethod
    def _get_standard_field_description(cls) -> str:
        return ('This class represents a concrete model for data items in the `omnipy` Python '
                'package. It is a statically typed specialization of the Model class, '
                'which is itself wrapping the excellent Python package named `pydantic`.')

    @classmethod
    def validate(cls: type['Model'], value: Any) -> 'Model':
        """
        Hack to allow overwriting of __iter__ method without compromising pydantic validation. Part
        of the pydantic API and not the Omnipy API.
        """
        # TODO: Doublecheck if validate() method is still needed for pydantic v2

        validate_cls_counts[cls.__name__] += 1
        if is_model_instance(value):

            @contextmanager
            def temporary_set_value_iter_to_pydantic_method() -> Iterator[None]:
                prev_iter = value.__class__.__iter__
                value.__class__.__iter__ = GenericModel.__iter__

                try:
                    yield
                finally:
                    value.__class__.__iter__ = prev_iter

            with temporary_set_value_iter_to_pydantic_method():
                return super().validate(value)
        else:
            return super().validate(value)

    @classmethod
    def update_forward_refs(cls, **localns: Any) -> None:
        prev_outer_type = cls._get_root_field().outer_type_
        prev_type = cls._get_root_field().type_

        super().update_forward_refs(**localns)

        calling_module = get_calling_module_name()
        cls._get_root_field().outer_type_ = evaluate_any_forward_refs_if_possible(
            prev_outer_type, calling_module, **localns)
        cls._get_root_field().type_ = evaluate_any_forward_refs_if_possible(
            prev_type, calling_module, **localns)
        cls.set_orig_model(
            evaluate_any_forward_refs_if_possible(cls.get_orig_model(), calling_module, **localns))
        cls.__annotations__[ROOT_KEY] = evaluate_any_forward_refs_if_possible(
            cls.__annotations__[ROOT_KEY], calling_module, **localns)

        cls._recursively_set_allow_none(cls._get_root_field())

        cls.__name__ = remove_forward_ref_notation(cls.__name__)
        cls.__qualname__ = remove_forward_ref_notation(cls.__qualname__)

        cls._clean_type_caches()

    def validate_contents(self) -> None:
        self._validate_and_set_value(self.contents)

    def _validate_and_set_value(
        self,
        new_contents: object,
        reset_solution: ContextManager[None] | None = None,
    ) -> None:

        old_contents_id = id(self.contents)

        def _set_new_contents(contents: object) -> None:
            if id(contents) != old_contents_id:
                self.contents = contents

        self._generic_validate_contents(
            new_contents=new_contents,
            outer_reset_solution=reset_solution,
            post_validation_func=_set_new_contents,
        )

    def _prepare_reset_solution_take_snapshot_if_needed(
        self,
        /,
    ) -> ResetSolutionTuple:
        snapshot_taken = False
        if self.config.interactive_mode:
            # TODO: Lazy snapshotting causes unneeded double validation for data that is later
            #       validated for snapshot. Perhaps add a dirty flag to snapshot that can be used
            #       to determine if re-validation is needed? This can also help avoid equality
            #       tests, which might be expensive for large data structures.
            needs_pre_validation = (not self.has_snapshot()
                                    or not self.contents_validated_according_to_snapshot())
            if needs_pre_validation:
                internal_reset_solution = self._get_reset_solution()
                with internal_reset_solution:
                    self._validate_and_set_value(
                        self.contents, reset_solution=internal_reset_solution)
                    snapshot_taken = True

        return ResetSolutionTuple(
            reset_solution=self._get_reset_solution(),
            snapshot_taken=snapshot_taken,
        )

    def _get_reset_solution(self) -> ContextManager[None]:
        if self.config.interactive_mode and self.has_snapshot():
            return self._get_revert_to_snapshot_reset_solution()
        else:
            return nothing()

    def _get_revert_to_snapshot_reset_solution(self) -> ContextManager[None]:
        prev_deepcopy_content_ids = SetDeque[int]()

        def _setup():
            prev_deepcopy_content_ids.extend(self.snapshot_holder.get_deepcopy_content_ids())

        def _handle_exception():
            new_deepcopy_content_ids = SetDeque[int](
                self.snapshot_holder.get_deepcopy_content_ids())
            new_deepcopy_content_ids.extend(prev_deepcopy_content_ids)
            self.snapshot_holder.schedule_deepcopy_content_ids_for_deletion(
                *new_deepcopy_content_ids)
            # self.contents = self.snapshot_holder.get_snapshot_deepcopy(self)
            from copy import deepcopy
            self.contents = deepcopy(self.snapshot)

        return setup_and_teardown_callback_context(
            setup_func=_setup,
            exception_func=_handle_exception,
        )

    def _generic_validate_contents(
        self,
        /,
        new_contents: object,
        outer_reset_solution: ContextManager[None] | None = None,
        post_validation_func: Callable[[_RootT], None] | None = None,
    ) -> None:
        keep_alive_old_contents = self.contents  # To ensure old content ids are not reused

        inner_reset_solution: ContextManager[None]
        if outer_reset_solution:
            inner_reset_solution = nothing()
        else:
            validating_self = new_contents is self.contents
            reset_solution_tuple = self._prepare_reset_solution_take_snapshot_if_needed()
            if validating_self and reset_solution_tuple.snapshot_taken:
                return
            inner_reset_solution = reset_solution_tuple.reset_solution

        with (inner_reset_solution):
            validated_contents = self._validate_contents_from_value(new_contents)

            if validated_contents is new_contents:
                validated_contents = copy(validated_contents)

            if post_validation_func:
                post_validation_func(validated_contents)
        del inner_reset_solution

        del new_contents
        self._take_snapshot_of_validated_contents()

        del keep_alive_old_contents

    def _validate_contents_from_value(
        self,
        value: object,
    ) -> _RootT:
        _is_model, value = self._prepare_value_for_validation_if_model(value)

        values, fields_set, validation_error = validate_model(self.__class__, {ROOT_KEY: value})
        if validation_error:
            raise validation_error

        return values[ROOT_KEY]

    @property
    def snapshot(self) -> _RootT:
        snapshot_wrapper = self._get_snapshot_wrapper()
        assert snapshot_wrapper.id == id(self)
        return snapshot_wrapper.snapshot

    def has_snapshot(self) -> bool:
        return self in self.snapshot_holder

    def _get_snapshot_wrapper(self) -> IsSnapshotWrapper['Model', _RootT]:
        assert self.has_snapshot(), 'No snapshot taken yet'
        return self.snapshot_holder[self]

    def snapshot_taken_of_same_model(self, model: 'Model') -> bool:
        snapshot_wrapper = self._get_snapshot_wrapper()
        return snapshot_wrapper.taken_of_same_obj(model)

    def snapshot_differs_from_model(self, model: 'Model') -> bool:
        snapshot_wrapper = self._get_snapshot_wrapper()
        return snapshot_wrapper.differs_from(model.contents)

    def contents_validated_according_to_snapshot(self) -> bool:
        needs_validation = self.snapshot_differs_from_model(self) \
            or not self.snapshot_taken_of_same_model(self)
        return not needs_validation

    def _take_snapshot_of_validated_contents(self) -> None:
        if self.config.interactive_mode:
            with self.deepcopy_context(self.snapshot_holder.take_snapshot_setup,
                                       self.snapshot_holder.take_snapshot_teardown):
                self.snapshot_holder.take_snapshot(self)

    @classmethod
    def _parse_data(cls, data: Any) -> _RootT:
        return data

    # TODO: Expand _generous_sequence_support to support iterators, such as dict_keys. Also see if
    #       it is possible to support general mappings in a similar way
    @root_validator(pre=True)
    def _generous_sequence_support(cls, root_obj: dict[str, _RootT | None]) -> Any:
        if ROOT_KEY in root_obj:
            value = root_obj[ROOT_KEY]
            outer_type = cls.outer_type()
            if lenient_issubclass(outer_type, Sequence) \
                    and not lenient_isinstance(value, outer_type) \
                    and isinstance(value, Sequence) \
                    and not sequence_like(value) \
                    and not any(isinstance(value, typ) for typ in (str, bytes)):
                return {ROOT_KEY: [_ for _ in value]}
        return root_obj

    @root_validator
    def _parse_root_object(cls, root_obj: dict[str, _RootT | None]) -> Any:
        assert ROOT_KEY in root_obj
        value = root_obj[ROOT_KEY]
        value = parse_none_according_to_model(value, root_model=cls)

        config = cls.data_class_creator.config
        with hold_and_reset_prev_attrib_value(config, 'dynamically_convert_elements_to_models'):
            config.dynamically_convert_elements_to_models = False
            return {ROOT_KEY: cls._parse_data(value)}

    @property
    def contents(self) -> _RootT:
        return cast(_RootT, self.__dict__.get(ROOT_KEY))

    @contents.setter
    def contents(self, value: _RootT) -> None:
        super().__setattr__(ROOT_KEY, value)

    def dict(self, *args, **kwargs) -> dict[str, object]:
        return {ROOT_KEY: self.to_data()}

    # TODO: Improve typing of to_data/from_data. Should be limited to JSON types at least, but also
    #       `_RootT` for simple models (without submodels). Handling Submodels is tricky, and may
    #       not be possible, e.g. `Model[list[Model[int]]().to_data()` should be type `list[int]`.
    #       A possibility is to support manually providing proper to_data, e.g. through
    #       Generic Mixin class, e.g.:
    #       ```python
    #       class MyModel(Model[list[Model[int]]], ModelData[list[int]]):
    #          ...
    #       ```
    def to_data(self) -> object:
        return super().dict(by_alias=True)[ROOT_KEY]

    def _empty_from_data(self, value: object) -> None:

        from contextlib import contextmanager

        @contextmanager
        def _reset_to_default(*args, **kwds):
            self.contents = self._get_default_value_from_model(self.full_type())
            yield

        self._validate_and_set_value(value, reset_solution=_reset_to_default())

    def from_data(self, value: object) -> None:
        if self.contents == self._get_default_value_from_model(self.full_type()):
            self._empty_from_data(value)
        else:
            self._validate_and_set_value(value)

    def absorb_and_replace(self, other: 'Model'):
        self.from_data(other.to_data())

    def to_json(self, pretty=True) -> str:
        json_content = self.json()
        if pretty:
            return self._pretty_print_json(json.loads(json_content))
        else:
            return json_content

    def from_json(self, json_contents: str) -> None:
        new_model = self.parse_raw(json_contents, proto=PydanticProtocol.json)
        self.contents = new_model.contents

    @classmethod
    @functools.cache
    def inner_type(cls, with_args: bool = False) -> TypeForm:
        return cls._get_root_type(outer=False, with_args=with_args)

    @classmethod
    @functools.cache
    def outer_type(cls, with_args: bool = False) -> TypeForm:
        return cls._get_root_type(outer=True, with_args=with_args)

    @classmethod
    @functools.cache
    def full_type(cls) -> TypeForm:
        return cls.outer_type(with_args=True)

    @classmethod
    @functools.cache
    def is_nested_type(cls) -> bool:
        return not cls.inner_type(with_args=True) == cls.outer_type(with_args=True)

    @classmethod
    def _clean_type_caches(cls):
        cls.outer_type.cache_clear()
        cls.inner_type.cache_clear()
        cls.full_type.cache_clear()
        cls.is_nested_type.cache_clear()

    @classmethod
    def _get_root_field(cls) -> ModelField:
        return cast(ModelField, cls.__fields__.get(ROOT_KEY))

    @classmethod
    def _get_root_type(cls, outer: bool, with_args: bool) -> TypeForm:
        root_field = cls._get_root_field()
        root_type = root_field.outer_type_ if outer else root_field.type_

        orig_model = cls.get_orig_model()
        if orig_model != Undefined:
            if not is_optional(root_type) and is_optional(orig_model):
                if outer:
                    root_type = Optional[root_type]
                else:
                    root_type = Optional[root_field.outer_type_]

        return root_type if with_args else ensure_plain_type(root_type)

    # @classmethod
    # def create_from_json(cls, data: str | tuple[str]):
    #     if isinstance(data, tuple):
    #         data = data[0]
    #
    #     obj = cls()
    #     obj.from_json(data)
    #     return obj
    #
    # def __reduce__(self):
    #     return self.__class__.create_from_json, (self.to_json(),)

    @classmethod
    def to_json_schema(cls, pretty=True) -> str:
        schema = cls.schema()
        if 'orig_model' in schema:
            del schema['orig_model']

        if pretty:
            return cls._pretty_print_json(schema)
        else:
            return json.dumps(schema)

    @staticmethod
    def _pretty_print_json(json_content: Any) -> str:
        return json.dumps(json_content, indent=2)

    def _check_for_root_key(self) -> None:
        if ROOT_KEY not in self.__dict__:
            raise TypeError('The Model class requires the specific model to be specified in as '
                            'a type hierarchy within brackets either directly, e.g.:\n'
                            '\t"model = Model[list[int]]([1,2,3])"\n'
                            'or indirectly in a subclass definition, e.g.:\n'
                            '\t"class MyNumberList(Model[list[int]]): ..."')

    def __setattr__(self, attr: str, value: Any) -> None:
        if attr in ['__module__'] + list(self.__dict__.keys()) and attr not in [ROOT_KEY]:
            super().__setattr__(attr, value)
        else:
            match (attr):
                case 'contents':
                    contents_prop = getattr(self.__class__, attr)
                    old_contents_id = id(contents_prop.__get__(self))
                    is_new_contents = id(value) != old_contents_id

                    if is_new_contents:
                        contents_prop.__set__(self, value)

                        if self.config.interactive_mode and self.has_snapshot():
                            self.snapshot_holder.schedule_deepcopy_content_ids_for_deletion(
                                old_contents_id)
                case 'repr_state':
                    prop = getattr(self.__class__, attr)
                    prop.__set__(self, value)
                case _:
                    if self._is_non_omnipy_pydantic_model():
                        self._special_method(
                            '__setattr__',
                            MethodInfo(state_changing=True, returns_same_type=YesNoMaybe.NO),
                            attr,
                            value)
                    else:
                        raise RuntimeError('Model does not allow setting of extra attributes')

    def _special_method(  # noqa: C901
            self, name: str, info: MethodInfo, *args: object, **kwargs: object) -> object:

        if info.state_changing and self.config.interactive_mode:

            def _call_special_method_and_return_self_if_inplace(*inner_args: object,
                                                                **inner_kwargs: object) -> object:
                return_val = self._call_special_method(name, *inner_args, **inner_kwargs)

                if id(return_val) == id(self.contents):  # in-place operator, e.g. model += 1
                    return_val = self

                return return_val

            reset_solution = self._prepare_reset_solution_take_snapshot_if_needed().reset_solution
            with reset_solution:
                ret = _call_special_method_and_return_self_if_inplace(*args, **kwargs)
                if ret is NotImplemented:
                    return ret

                self._validate_and_set_value(
                    new_contents=self.contents,
                    reset_solution=reset_solution,
                )

        elif name == '__iter__' and isinstance(self, Iterable):
            _per_element_model_generator = self._get_convert_full_element_model_generator(
                cast(Iterable, self.contents),
                level_up_type_arg_idx=0,
            )
            return _per_element_model_generator()
        else:
            ret = self._call_special_method(name, *args, **kwargs)
            if ret is NotImplemented:
                return ret

            if info.state_changing:
                self.validate_contents()

        if id(ret) != id(self) and info.returns_same_type:
            level_up = False
            if name == '__getitem__':
                assert len(args) == 1
                if not isinstance(args[0], slice):
                    level_up = True

            # We can do this with some ease of mind as all the methods except '__getitem__' with
            # integer argument are supposed to possibly return a result of the same type.
            ret = self._convert_to_model_if_reasonable(
                ret,
                level_up=level_up,
                level_up_arg_idx=-1,
                raise_validation_errors=(info.returns_same_type == YesNoMaybe.YES),
            )

        return ret

    def _call_special_method(  # noqa: C901
            self,
            name: str,
            *args: object,
            **kwargs: object,
    ) -> object:
        contents = self._get_real_contents()
        has_add_method = hasattr(contents, '__add__')
        has_radd_method = hasattr(contents, '__radd__')
        has_iadd_method = hasattr(contents, '__iadd__')

        if name == '__add__' and has_add_method:

            def _add(other):
                # try:
                #     return contents.__add__(self.__class__(other).contents)
                # except ValidationError:
                return contents.__add__(other)

            # return _add_new_other_model(*args, **kwargs)
            method = _add
            return self._call_single_arg_method_with_model_converted_other_first(
                name, method, *args, **kwargs)

        elif name == '__radd__' and (has_radd_method or has_add_method):

            def _radd(other):
                if has_radd_method:
                    return contents.__radd__(other)
                else:
                    return contents.__add__(other)

            def _radd_model_converted_other(other):
                if has_radd_method:
                    return contents.__radd__(self.__class__(other).contents)
                else:
                    return self.__class__(other).contents.__add__(self.contents)

            method = _radd
            model_converted_other_method = _radd_model_converted_other
            return self._call_single_arg_method_with_model_converted_other_first(
                name,
                method,
                *args,
                model_converted_other_method=model_converted_other_method,
                **kwargs,
            )

        elif name == '__iadd__' and (has_iadd_method or has_add_method):

            def _iadd(other):
                if has_iadd_method:
                    return contents.__iadd__(other)
                else:
                    return contents.__add__(other)

            method = _iadd
            return self._call_single_arg_method_with_model_converted_other_first(
                name, method, *args, **kwargs)
        else:
            try:
                method = cast(Callable, self._getattr_from_contents_obj(name))
            except AttributeError as e:
                if name in ('__int__', '__float__', '__complex__'):
                    raise ValueError from e
                if name == '__len__':
                    raise TypeError(f"object of type '{self.__class__.__name__}' has no len()")
                else:
                    return NotImplemented

            if name == '__hash__' and method is None:
                raise TypeError(f'unhashable type: {self.__class__.__name__}')

            return self._call_method_with_unconverted_args_first(method, *args, **kwargs)

    def _call_single_arg_method_with_model_converted_other_first(
        self,
        name: str,
        method: Callable,
        *args: object,
        model_converted_other_method: Callable | None = None,
        **kwargs: object,
    ):
        if len(args) != 1:
            raise TypeError(f'expected 1 argument, got {len(args)}')

        if len(kwargs) > 0:
            raise TypeError(f'method {name} takes no keyword arguments')

        arg = args[0]

        try:
            try:
                if model_converted_other_method:
                    return model_converted_other_method(arg)
                else:
                    return method(self.__class__(arg).contents, **kwargs)
            except ValidationError:
                # TODO: Add debug logging for hidden validation and other exceptions e.g. when
                #       concatenating `Model[int](123) + '234.'` (gives TypeError:
                #       unsupported operand type(s) for +: 'Model[int]' and 'str'). ?
                #       `Model[int](123) + '234'` works fine, returns `Model[int]('357')`.
                return method(arg)
        except TypeError:
            return NotImplemented

    def _call_method_with_unconverted_args_first(
        self,
        method: Callable,
        *args: object,
        **kwargs: object,
    ):
        try:
            with hold_and_reset_prev_attrib_value(
                    self.config,
                    'dynamically_convert_elements_to_models',
            ):
                self.config.dynamically_convert_elements_to_models = False
                ret = method(*args, **kwargs)
        except TypeError as type_exc:
            try:
                ret = self._call_method_with_model_converted_args(method, *args, **kwargs)
            except ValidationError:
                raise type_exc
        if ret is NotImplemented:
            try:
                ret = self._call_method_with_model_converted_args(method, *args, **kwargs)
            except ValidationError:
                pass
        return ret

    def _call_method_with_model_converted_args(
        self,
        method: Callable,
        *args: object,
        **kwargs: object,
    ):
        model_args = [self.__class__(arg).contents for arg in args]
        return method(*model_args, **kwargs)

    def _get_convert_full_element_model_generator(
            self, elements: Iterable | None,
            level_up_type_arg_idx: int | slice) -> Callable[..., Generator]:
        def _convert_full_element_model_generator(elements=elements):
            for el in elements:
                yield self._convert_to_model_if_reasonable(
                    el,
                    level_up=True,
                    level_up_arg_idx=level_up_type_arg_idx,
                )

        return _convert_full_element_model_generator

    def _get_convert_element_value_model_generator(
            self, elements: Iterable | None) -> Callable[..., Generator]:
        def _convert_element_value_model_generator(elements=elements):
            for el in elements:
                yield (
                    el[0],
                    self._convert_to_model_if_reasonable(
                        el[1],
                        level_up=True,
                        level_up_arg_idx=1,
                    ),
                )

        return _convert_element_value_model_generator

    def _convert_to_model_if_reasonable(  # noqa: C901
        self,
        ret: Mapping[_KeyT, _ValT] | Iterable[_ValT] | _ReturnT | _RootT,
        level_up: bool = False,
        level_up_arg_idx: int = 1,
        raise_validation_errors: bool = False,
    ) -> ('Model[_KeyT] | Model[_ValT] | Model[tuple[_KeyT, _ValT]] '
          '| Model[_ReturnT] | Model[_RootT] | _ReturnT'):

        if level_up and not self.config.dynamically_convert_elements_to_models:
            ...
        elif is_model_instance(ret):
            ...
        else:
            outer_type = self.outer_type(with_args=True)
            # For double Models, e.g. Model[Model[int]], where _get_real_contents() have already
            # skipped the outer Model to get the `ret`, we need to do the same to compare the value
            # with the corresponding type.
            if lenient_issubclass(ensure_plain_type(outer_type), Model):
                outer_type = cast(Model, outer_type).outer_type(with_args=True)

            for type_to_check in all_type_variants(outer_type):
                plain_type_to_check = ensure_plain_type(type_to_check)
                if plain_type_to_check in (ForwardRef, TypeVar, None):
                    continue

                if level_up:
                    type_args = get_args(type_to_check)
                    if len(type_args) == 0:
                        type_args = (type_to_check,)
                    if type_args:
                        for level_up_type_to_check in all_type_variants(
                                type_args[level_up_arg_idx]):
                            level_up_type_to_check = self._fix_tuple_type_from_args(
                                level_up_type_to_check)
                            if self._is_instance_or_literal(
                                    ret,
                                    ensure_plain_type(level_up_type_to_check),
                                    level_up_type_to_check,
                            ):
                                try:
                                    return Model[level_up_type_to_check](ret)  # type: ignore
                                except ValidationError:
                                    if raise_validation_errors:
                                        raise
                                except TypeError:
                                    pass

                else:
                    if self._is_instance_or_literal(
                            ret,
                            plain_type_to_check,
                            type_to_check,
                    ):
                        try:
                            return self.__class__(ret)
                        except ValidationError:
                            if raise_validation_errors:
                                raise
                        except TypeError:
                            pass

        return cast(_ReturnT, ret)

    @staticmethod
    def _is_instance_or_literal(obj: object, plain_type: type, raw_type: type | GenericAlias):
        if plain_type is Literal:
            args = get_args(raw_type)
            for arg in args:
                if obj == arg:
                    return True
            return False
        else:
            return lenient_isinstance(obj, plain_type)

    def _fix_tuple_type_from_args(
        self, level_up_type_to_check: type | GenericAlias | tuple[type | GenericAlias, ...]
    ) -> type | GenericAlias:
        if isinstance(level_up_type_to_check, tuple):
            match len(level_up_type_to_check):
                case 1:
                    return level_up_type_to_check[0]
                case _:
                    return tuple[level_up_type_to_check]  # type: ignore[valid-type]
        else:
            return level_up_type_to_check

    def __getattr__(self, attr: str) -> Any:
        if self._is_non_omnipy_pydantic_model() and self._contents_obj_hasattr(attr):
            self._validate_and_set_value(self.contents)

        contents_attr = self._getattr_from_contents_obj(attr)

        if inspect.isroutine(contents_attr):
            reset_solution = self._prepare_reset_solution_take_snapshot_if_needed().reset_solution
            new_contents_attr = self._getattr_from_contents_obj(attr)

            def _validate_contents(ret: Any):
                self._validate_and_set_value(self.contents, reset_solution=reset_solution)
                return ret

            contents_attr = add_callback_after_call(new_contents_attr,
                                                    _validate_contents,
                                                    reset_solution)

        if attr in ('values', 'items'):
            match attr:
                case 'values':
                    _model_generator = self._get_convert_full_element_model_generator(
                        None,
                        level_up_type_arg_idx=1,
                    )
                case 'items':
                    _model_generator = self._get_convert_element_value_model_generator(None,)

            contents_attr = add_callback_after_call(contents_attr, _model_generator, no_context)

        return contents_attr

    def _is_non_omnipy_pydantic_model(self) -> bool:
        return is_non_omnipy_pydantic_model(self._get_real_contents())

    def _contents_obj_hasattr(self, attr) -> object:
        return hasattr(self._get_real_contents(), attr)

    def _getattr_from_contents_obj(self, attr) -> object:
        return getattr(self._get_real_contents(), attr)

    def _getattr_from_contents_cls(self, attr) -> object:
        return getattr(self._get_real_contents().__class__, attr)

    def _get_real_contents(self) -> object:
        if is_model_instance(self.contents):
            return self.contents.contents
        else:
            return self.contents

    def __eq__(self, other: object) -> bool:
        return is_model_instance(other) \
            and self.__class__ == other.__class__ \
            and all_equals(self.contents, cast(Model, other).contents)
        # and self.to_data() == cast(Model, other).to_data()  # last line is just in case

    def __bool__(self):
        if self._get_real_contents():
            return True
        else:
            return False

    def __call__(self, *args: object, **kwargs: object) -> object:
        if not hasattr(self._get_real_contents(), '__call__'):
            raise TypeError(f"'{self.__class__.__name__}' object is not callable")
        return self._special_method(
            '__call__',
            MethodInfo(state_changing=True, returns_same_type=YesNoMaybe.NO),
            *args,
            **kwargs)

    def __repr_args__(self):
        return [(None, self.contents)]



from collections import defaultdict
from datetime import datetime
from typing import DefaultDict

from omnipy.api.enums import RunState, RunStateLogMessages
from omnipy.api.protocols.private.compute.mixins import IsUniquelyNamedJob
from omnipy.hub.log.mixin import LogMixin


class RunStateRegistry(LogMixin):
    def __init__(self) -> None:

        self._jobs: dict[str, IsUniquelyNamedJob] = {}
        self._job_states: dict[str, RunState] = {}
        self._state_jobs: DefaultDict[RunState, list[str]] = defaultdict(list)
        self._job_state_datetime: dict[tuple[str, RunState], datetime] = {}

        super().__init__()

    def get_job_state(self, job: IsUniquelyNamedJob) -> RunState:
        return self._job_states[job.unique_name]

    def get_job_state_datetime(self, job: IsUniquelyNamedJob, state: RunState) -> datetime:
        return self._job_state_datetime[(job.unique_name, state)]

    def all_jobs(self, state: RunState | None = None) -> tuple[IsUniquelyNamedJob, ...]:
        if state is not None:
            job_unique_names = self._state_jobs[state]
            return tuple(self._jobs[unique_name] for unique_name in job_unique_names)
        else:
            return tuple(self._jobs.values())

    def set_job_state(self, job: IsUniquelyNamedJob, state: RunState) -> None:
        cur_datetime = datetime.now()

        if job.unique_name in self._jobs:
            self._update_job_registration(job, state)
        else:
            self._register_new_job(job, state)

        self._update_job_stats(job, state, cur_datetime)
        self._log_state_change(job, state)

    def _other_job_registered_with_same_unique_name(self, job: IsUniquelyNamedJob) -> bool:
        other_job_same_unique_name = self._jobs.get(job.unique_name)
        return bool(other_job_same_unique_name) and id(other_job_same_unique_name) != id(job)

    def _update_job_registration(self, job: IsUniquelyNamedJob, state: RunState) -> None:
        # TODO: Reimplement logic using a state machine, e.g. "transitions" package
        if self._other_job_registered_with_same_unique_name(job):
            while self._other_job_registered_with_same_unique_name(job):
                job.regenerate_unique_name()
            self._register_new_job(job, state)
        else:
            prev_state = self._job_states[job.unique_name]
            if state == prev_state + 1:
                self._state_jobs[prev_state].remove(job.unique_name)
            else:
                self._raise_job_error(
                    job,
                    f'Transitioning from state {prev_state.name} '
                    f'to state {state.name} is not allowed',
                )

    def _register_new_job(self, job, state) -> None:
        if state != RunState.INITIALIZED:
            self._raise_job_error(
                job,
                f'Initial state of must be "INITIALIZED", not "{state.name}"',
            )
        self._jobs[job.unique_name] = job

    def _update_job_stats(self, job, state, cur_datetime) -> None:
        self._job_states[job.unique_name] = state
        self._state_jobs[state].append(job.unique_name)
        self._job_state_datetime[(job.unique_name, state)] = cur_datetime

    def _log_state_change(self, job: IsUniquelyNamedJob, state: RunState) -> None:
        log_msg = RunStateLogMessages[state.name].value.format(job.unique_name)
        datetime_obj = self.get_job_state_datetime(job, state)
        self.log(log_msg, datetime_obj=datetime_obj)

    def _raise_job_error(self, job: IsUniquelyNamedJob, msg: str) -> None:
        raise ValueError(f'Error in job "{job.unique_name}": {msg}')



from pydantic import NoneIsNotAllowedError


class JobStateException(Exception):
    ...


class ShouldNotOccurException(Exception):
    ...


class OmnipyNoneIsNotAllowedError(NoneIsNotAllowedError):
    msg_template = '[Omnipy] none is not an allowed value'


class PendingDataError(Exception):
    ...


class FailedDataError(Exception):
    ...



from datetime import date, datetime
from typing import Union

from pydantic import constr

# TODO: Remove date/datetime validator with pydantic v2, `model_dump(mode='json')` should fix it
#       (see https://github.com/pydantic/pydantic/issues/1409#issuecomment-1423995424).


def date_to_iso_format(data: Union[datetime, date, constr(max_length=0)]):
    if isinstance(data, date) or isinstance(data, datetime):
        return data.isoformat()
    return data



from abc import ABC, abstractmethod
from io import BytesIO
import os
import tarfile
from tarfile import TarInfo
from typing import Any, Callable, IO, Type

from pydantic import ValidationError

from omnipy.api.protocols.private.log import CanLog
from omnipy.api.protocols.public.data import IsDataset, IsSerializer, IsTarFileSerializer
from omnipy.util.contexts import hold_and_reset_prev_attrib_value


class Serializer(ABC):
    """"""
    @classmethod
    @abstractmethod
    def is_dataset_directly_supported(cls, dataset: IsDataset) -> bool:
        pass

    @classmethod
    @abstractmethod
    def get_dataset_cls_for_new(cls) -> Type[IsDataset]:
        pass

    @classmethod
    @abstractmethod
    def get_output_file_suffix(cls) -> str:
        pass

    @classmethod
    @abstractmethod
    def serialize(cls, dataset: IsDataset) -> bytes | memoryview:
        pass

    @classmethod
    @abstractmethod
    def deserialize(cls, serialized: bytes, any_file_suffix=False) -> IsDataset:
        pass


class TarFileSerializer(Serializer, ABC):
    """"""
    @classmethod
    def create_tarfile_from_dataset(cls,
                                    dataset: IsDataset,
                                    data_encode_func: Callable[[Any], bytes | memoryview]):
        bytes_io = BytesIO()
        with tarfile.open(fileobj=bytes_io, mode='w:gz') as tarfile_stream:
            for data_file, data in dataset.items():
                json_data_bytestream = BytesIO(data_encode_func(data.contents))
                json_data_bytestream.seek(0)
                tarinfo = TarInfo(name=f'{data_file}.{cls.get_output_file_suffix()}')
                tarinfo.size = len(json_data_bytestream.getbuffer())
                tarfile_stream.addfile(tarinfo, json_data_bytestream)
        return bytes_io.getbuffer().tobytes()

    @classmethod
    def create_dataset_from_tarfile(cls,
                                    dataset: IsDataset,
                                    tarfile_bytes: bytes,
                                    data_decode_func: Callable[[IO[bytes]], Any],
                                    dictify_object_func: Callable[[str, Any], dict | str],
                                    import_method: str = 'from_data',
                                    any_file_suffix: bool = False):
        with tarfile.open(fileobj=BytesIO(tarfile_bytes), mode='r:gz') as tarfile_stream:
            for filename in tarfile_stream.getnames():
                data_file = tarfile_stream.extractfile(filename)
                if not any_file_suffix:
                    assert filename.endswith(f'.{cls.get_output_file_suffix()}')
                data_file_name = os.path.basename('.'.join(filename.split('.')[:-1]))
                getattr(dataset, import_method)(
                    dictify_object_func(data_file_name, data_decode_func(data_file)))


class SerializerRegistry:
    def __init__(self) -> None:
        self._serializer_classes: list[Type[IsSerializer]] = []

    def register(self, serializer_cls: Type[IsSerializer]) -> None:
        self._serializer_classes.append(serializer_cls)

    @property
    def serializers(self) -> tuple[Type[IsSerializer], ...]:
        return tuple(self._serializer_classes)

    @property
    def tar_file_serializers(self) -> tuple[Type[IsTarFileSerializer], ...]:
        return tuple(cls for cls in self._serializer_classes if issubclass(cls, TarFileSerializer))

    def auto_detect(self, dataset: IsDataset):
        return self._autodetect_serializer(dataset, self.serializers)

    def auto_detect_tar_file_serializer(self, dataset: IsDataset):
        return self._autodetect_serializer(dataset, self.tar_file_serializers)

    @classmethod
    def _autodetect_serializer(
        cls,
        dataset: IsDataset,
        serializers: tuple[Type[IsSerializer], ...],
    ) -> tuple[IsDataset, IsSerializer]:

        from omnipy.hub.runtime import runtime
        if runtime:
            with hold_and_reset_prev_attrib_value(
                    runtime.config.data,
                    'interactive_mode',
            ):
                with hold_and_reset_prev_attrib_value(
                        runtime.config.data,
                        'dynamically_convert_elements_to_models',
                ):
                    runtime.config.data.interactive_mode = False
                    runtime.config.data.dynamically_convert_elements_to_models = False

                    return cls._test_all_serializer_combos(dataset, serializers)
        else:
            return cls._test_all_serializer_combos(dataset, serializers)

    @classmethod
    def _test_all_serializer_combos(
        self,
        dataset: IsDataset,
        serializers: tuple[Type[IsSerializer], ...],
    ) -> tuple[IsDataset, IsSerializer]:

        # def _direct(dataset: Dataset, serializer: Serializer):
        #     new_dataset_cls = serializer.get_dataset_cls_for_new()
        #     new_dataset = new_dataset_cls(dataset)
        #     return new_dataset

        def _to_data_from_json(dataset: IsDataset, serializer: IsSerializer):
            new_dataset_cls = serializer.get_dataset_cls_for_new()
            new_dataset = new_dataset_cls()
            new_dataset.from_json(dataset.to_data())
            return new_dataset

        def _to_data_from_data(dataset: IsDataset, serializer: IsSerializer):
            new_dataset_cls = serializer.get_dataset_cls_for_new()
            new_dataset = new_dataset_cls()
            new_dataset.from_data(dataset.to_data())
            return new_dataset

        def _to_data_from_data_if_direct(dataset: IsDataset, serializer: IsSerializer):
            assert serializer.is_dataset_directly_supported(dataset)
            return _to_data_from_data(dataset, serializer)

        # def _to_json_from_json(dataset: Dataset, serializer: Serializer):
        #     new_dataset_cls = serializer.get_dataset_cls_for_new()
        #     new_dataset = new_dataset_cls()
        #     new_dataset.from_json(dataset.to_json())
        #     return new_dataset

        for func in (_to_data_from_data_if_direct, _to_data_from_json, _to_data_from_data):
            for serializer in serializers:
                try:
                    new_dataset = func(dataset, serializer)
                    return new_dataset, serializer
                except (TypeError, ValueError, ValidationError, AssertionError):
                    pass

        return None, None

    def detect_tar_file_serializers_from_dataset_cls(self, dataset: IsDataset):
        serializers = tuple(
            serializer_cls for serializer_cls in self.tar_file_serializers
            if serializer_cls.is_dataset_directly_supported(dataset))
        if len(serializers) == 0:
            serializers = tuple(serializer_cls for serializer_cls in self.tar_file_serializers
                                if serializer_cls.get_output_file_suffix() == 'bytes')
        return serializers

    def detect_tar_file_serializers_from_file_suffix(self, file_suffix: str):
        return tuple(serializer_cls for serializer_cls in self.tar_file_serializers
                     if serializer_cls.get_output_file_suffix() == file_suffix)

    def load_from_tar_file_path_based_on_file_suffix(self,
                                                     log_obj: CanLog,
                                                     tar_file_path: str,
                                                     to_dataset: IsDataset):
        if hasattr(log_obj, 'log'):
            log = log_obj.log
        else:
            log = print

        with tarfile.open(tar_file_path, 'r:gz') as tarfile_obj:
            file_suffixes = set(fn.split('.')[-1] for fn in tarfile_obj.getnames())
        if len(file_suffixes) != 1:
            log(f'Tar archive contains files with different or '
                f'no file suffixes: {file_suffixes}. Serializer '
                f'cannot be uniquely determined. Aborting '
                f'restore.')
        else:
            file_suffix = file_suffixes.pop()
            serializers = self.detect_tar_file_serializers_from_file_suffix(file_suffix)
            if len(serializers) == 0:
                log(f'No serializer for file suffix "{file_suffix}" can be'
                    f'determined. Aborting restore.')
            else:
                log(f'Reading dataset from a gzipped tarpack at'
                    f' "{os.path.abspath(tar_file_path)}"')

                serializer = serializers[0]
                with open(tar_file_path, 'rb') as tarfile_binary:
                    auto_dataset = serializer.deserialize(tarfile_binary.read())

                if to_dataset.get_model_class() is auto_dataset.get_model_class():
                    to_dataset.data = auto_dataset.data
                    return to_dataset
                else:
                    try:
                        if to_dataset.get_model_class().inner_type == str:
                            to_dataset.from_data(auto_dataset.to_json())
                        else:
                            to_dataset.from_json(auto_dataset.to_data())
                        return to_dataset
                    except Exception:
                        return auto_dataset

    def load_from_tar_file_path_based_on_dataset_cls(self,
                                                     log_obj: CanLog,
                                                     tar_file_path: str,
                                                     to_dataset: IsDataset):
        if hasattr(log_obj, 'log'):
            log = log_obj.log
        else:
            log = print

        serializers = self.detect_tar_file_serializers_from_dataset_cls(to_dataset)
        if len(serializers) == 0:
            log(f'No serializer for Dataset with type "{type(to_dataset)}" can be '
                f'determined.')
        else:
            for serializer in serializers:
                log(f'Reading dataset from a gzipped tarpack at'
                    f' "{os.path.abspath(tar_file_path)}" with serializer type: '
                    f'"{serializer.__name__}"')

                with open(tar_file_path, 'rb') as tarfile_binary:
                    out_dataset = serializer.deserialize(tarfile_binary.read(), any)

                return out_dataset



from itertools import chain
from typing import Any, get_args, Mapping, MutableSequence

from pydantic.fields import Undefined
from pydantic.typing import is_none_type
from pydantic.utils import lenient_isinstance, lenient_issubclass
from typing_extensions import TypeVar

from omnipy.api.exceptions import OmnipyNoneIsNotAllowedError
from omnipy.api.typedefs import TypeForm
from omnipy.data.helpers import is_model_subclass
from omnipy.util.helpers import is_optional, is_union

_RootT = TypeVar('_RootT')


# Partial workaround of https://github.com/pydantic/pydantic/issues/3836 and similar bugs,
# together with hacks setting allow_none=True (_ModelMetaclass and _recursively_set_allow_none).
# See series of relevant tests in test_model.py starting with  test_list_of_none_variants().
def parse_none_according_to_model(value: _RootT, root_model) -> _RootT:  # IsModel
    outer_type = root_model.outer_type(with_args=True)
    plain_outer_type = root_model.outer_type(with_args=False)
    outer_args = get_args(outer_type)

    if is_model_subclass(outer_type):
        return _parse_none_in_model(outer_type, value)

    if root_model.is_nested_type():
        inner_val_type = root_model.inner_type(with_args=True)

        # Mutable sequences or variable tuples
        if _outer_type_and_value_are_of_types(plain_outer_type, value, (MutableSequence, tuple)):
            return _parse_none_in_mutable_sequence_or_tuple(plain_outer_type, inner_val_type, value)

        # Mappings
        if _outer_type_and_value_are_of_types(plain_outer_type, value, Mapping) and outer_args:
            return _parse_none_in_mapping(plain_outer_type, outer_args, inner_val_type, value)

        if lenient_isinstance(outer_type, TypeVar):
            return _parse_none_in_typevar(inner_val_type)

        if value is None:
            raise OmnipyNoneIsNotAllowedError()

    else:
        union_variants = _split_outer_type_to_union_variants(outer_args)
        flattened_union_variants = _flatten_two_level_tuple(union_variants)

        if any(is_model_subclass(tp_) or _supports_none(tp_) for tp_ in flattened_union_variants):

            # Fixed tuples
            if _outer_type_and_value_are_of_types(plain_outer_type, value, tuple) and outer_args:
                return _parse_none_in_fixed_tuple(plain_outer_type, union_variants, value)

            # Unions
            if is_union(plain_outer_type):
                return _parse_none_in_union(flattened_union_variants, value)

    return value


def _parse_none_in_model(outer_type, value):
    return outer_type(value)


def _split_to_union_variants(type_: TypeForm) -> tuple[TypeForm]:
    return get_args(type_) if is_union(type_) else (type_,)


def _split_outer_type_to_union_variants(outer_type_args):
    return tuple(_split_to_union_variants(_) for _ in outer_type_args)


def _flatten_two_level_tuple(two_level_tuple):
    return tuple(el for first_level_tuple in two_level_tuple for el in first_level_tuple)


def _supports_none(type_: TypeForm) -> bool:
    return is_none_type(type_) or is_optional(type_) or type_ in (object, Any)


def _outer_type_and_value_are_of_types(plain_outer_type, value, *types):
    return any(
        lenient_issubclass(plain_outer_type, type_) and lenient_isinstance(value, type_)
        for type_ in types)


def _parse_none_in_mutable_sequence_or_tuple(plain_outer_type, inner_val_type, value):
    inner_val_union_types = _split_to_union_variants(inner_val_type)

    if any(is_model_subclass(_) or _supports_none(_) for _ in inner_val_union_types):
        return plain_outer_type(
            _parse_none_in_types(inner_val_union_types) if val is None else val for val in value)
    return value


def _parse_none_in_mapping(plain_outer_type, outer_type_args, inner_val_type, value):
    inner_val_union_types = _split_to_union_variants(inner_val_type)

    inner_key_type = outer_type_args[0] if lenient_issubclass(
        plain_outer_type, Mapping) and outer_type_args else Undefined
    inner_key_union_types = _split_to_union_variants(inner_key_type)

    if any(
            is_model_subclass(_) or _supports_none(_)
            for _ in chain(inner_key_union_types, inner_val_union_types)):
        return plain_outer_type({
            _parse_none_in_types(inner_key_union_types) if key is None else key:
                _parse_none_in_types(inner_val_union_types) if val is None else val for key,
            val in value.items()
        })
    return value


def _parse_none_in_typevar(inner_val_type):
    inner_val_union_types = _split_to_union_variants(inner_val_type)

    return _parse_none_in_types(inner_val_union_types)


def _parse_none_in_fixed_tuple(plain_outer_type, tuple_of_union_variant_types, value):
    return plain_outer_type(
        _parse_none_in_types(tuple_of_union_variant_types[i]) if val is None else val for i,
        val in enumerate(value))


def _parse_none_in_union(flattened_union_variant_types, value):
    if value is None:
        return _parse_none_in_types(flattened_union_variant_types)
    else:
        return value


def _parse_none_in_types(inner_union_types: tuple[TypeForm]) -> object:
    for type_ in inner_union_types:
        if is_model_subclass(type_):
            return type_(parse_none_according_to_model(None, type_))
        elif _supports_none(type_):
            return None
    raise OmnipyNoneIsNotAllowedError()



from types import MappingProxyType
from typing import Generic, TypeAlias

from typing_extensions import ForwardRef

from omnipy.data.helpers import DoubleTypeVarStore, TypeVarStore
from omnipy.data.model import Model
from omnipy.modules.frozen.typedefs import FrozenDict, KeyT, ValT
from omnipy.modules.general.models import NotIterableExceptStrOrBytesModel

# TODO: Follow pydantic topic https://github.com/pydantic/pydantic/issues/6868 on MappingProxyType.
#       Used way too much energy to implement (and test) recursive frozen models, only to discover
#       that pydantic actually does not support MappingProxyType, which was hugely surprising.
#       Add tests on the immutability part once there is support for this.

# Note: see comments in omnipy.modules.json.models for more information on the basic organisation of
#       nested models.

# Note 2: currently crashes mypy (v1.10), probably due to recursive type aliases combined with
#         TypeVars from typing_extensions (with default). Not urgent to fix, as this in any case
#         awaits pydantic 2 and issue 6868 to be resolved. Keep out of __all__ until mypy issue
#         is fixed in order to avoid crashes with pytest-mypy-plugins, e.g. "test_json_types.yml".

#
# Private models
#


class _FrozenScalarM(Model[ValT], Generic[ValT]):
    _parse_data = NotIterableExceptStrOrBytesModel._parse_data


class _FrozenScalarWithTwoParamsM(Model[ValT], Generic[KeyT, ValT]):
    ...


# class _FrozenTupleBaseM(Model[tuple[_FrozenBaseT, ...]], Generic[_FrozenBaseT]):
#     ...

# class _FrozenDictBaseM(Model[FrozenDict[KeyT, _FrozenBaseT]], Generic[KeyT, _FrozenBaseT]):
#     ...


class _FrozenTupleM(Model[tuple[DoubleTypeVarStore[KeyT, ValT]
                                | ForwardRef('_FrozenAnyUnion[KeyT, ValT]'),
                                ...]],
                    Generic[KeyT, ValT]):
    ...


#
# class _FrozenDictM(_FrozenDictBaseM[KeyT, '_FrozenAnyUnion'], Generic[KeyT, ValT]):
#     ...


class _FrozenDictM(Model[FrozenDict[KeyT, TypeVarStore[ValT]
                                    | ForwardRef('_FrozenAnyUnion')]],
                   Generic[KeyT, ValT]):
    def to_data(self) -> dict[KeyT, ValT]:
        to_data = super().to_data()
        return MappingProxyType({key: val.to_data() for key, val in to_data.items()})


class _FrozenNoDictsM(Model[tuple[TypeVarStore[ValT]
                                  | ForwardRef('_FrozenNoDictsUnion[ValT]'), ...]],
                      Generic[ValT]):
    ...


# class _FrozenNoTuplesM(_FrozenDictBaseM['KeyT', '_FrozenNoTuplesUnion'], Generic[KeyT, ValT]):
#     ...


class _FrozenNoTuplesM(Model[FrozenDict[KeyT,
                                        DoubleTypeVarStore[KeyT, ValT]
                                        | ForwardRef('_FrozenNoTuplesUnion')]],
                       Generic[KeyT, ValT]):
    def to_data(self) -> dict[KeyT, ValT]:
        to_data = super().to_data()
        return MappingProxyType({key: val.to_data() for key, val in to_data.items()})


# TypeAliases

_FrozenAnyUnion: TypeAlias = \
    DoubleTypeVarStore[KeyT, ValT] | _FrozenScalarM[ValT] \
    | _FrozenTupleM[KeyT, ValT] | _FrozenDictM[KeyT, ValT]
_FrozenTupleM.update_forward_refs()
_FrozenDictM.update_forward_refs()
_FrozenNoDictsUnion: TypeAlias = TypeVarStore[ValT] | _FrozenScalarM[ValT] | _FrozenNoDictsM[ValT]
_FrozenNoTuplesUnion: TypeAlias = DoubleTypeVarStore[
    KeyT, ValT] | _FrozenScalarM[ValT] | _FrozenNoTuplesM[KeyT, ValT]


class _FrozenAnyUnionM(Model[_FrozenAnyUnion[KeyT, ValT]], Generic[KeyT, ValT]):
    ...


# Basic models needs to update their forward_refs with type aliases declared above

_FrozenTupleM.update_forward_refs()
_FrozenDictM.update_forward_refs()
_FrozenAnyUnionM.update_forward_refs()
_FrozenNoDictsM.update_forward_refs()
_FrozenNoTuplesM.update_forward_refs()

#
# Exportable models
#


class NestedFrozenDictsOrTuplesModel(Model[_FrozenAnyUnionM[KeyT, ValT]], Generic[KeyT, ValT]):
    """
    Recursive model for nested immutable containers (FrozenDict and tuples). Not functional.

    Awaiting support for MappingProxyType in pydantic for the immutability to actually work.
    Also use as generic types awaits Python 3.13. Way ahead of the crowd here!
    """


class NestedFrozenOnlyTuplesModel(Model[_FrozenNoDictsM[ValT]], Generic[ValT]):
    """
    Recursive model for nested tuples.

    Awaiting support for MappingProxyType in pydantic for the immutability to actually work.
    Also use as generic types awaits Python 3.13. Way ahead of the crowd here!
    """


class NestedFrozenOnlyDictsModel(Model[_FrozenNoTuplesM[KeyT, ValT]], Generic[KeyT, ValT]):
    """
    Recursive model for nested FrozenDicts.

    Awaiting support for MappingProxyType in pydantic for the immutability to actually work.
    Also use as generic types awaits Python 3.13. Way ahead of the crowd here!
    """



#!/usr/bin/env python

import json

import requests


def get_from_filter():
    for endpt_str in ['projects', 'cases', 'files', 'annotations']:
        endpt = 'https://api.gdc.cancer.gov/' + endpt_str
        endpt_map = endpt + '/_mapping'
        response = requests.get(endpt_map)
        fields = response.json()['fields']
        fields = ','.join(fields)

        with open(endpt_str + '_filter.json', 'r') as infile:
            filters = json.load(infile)
            params = {  # 'fields': fields,
                'filters': json.dumps(filters),
            }
        response = requests.get(endpt, params=params)
        print(endpt_str, 'status code', response.status_code)
        with open(endpt_str + '.json', 'w') as outfile:
            json.dump(response.json()['data'], outfile, indent=4)



from types import GenericAlias, UnionType
from typing import _AnnotatedAlias  # type: ignore[attr-defined]
from typing import _LiteralGenericAlias  # type: ignore[attr-defined]
from typing import _UnionGenericAlias  # type: ignore[attr-defined]
from typing import _SpecialForm, Callable, ForwardRef, TypeAlias, TypeVar

GeneralDecorator = Callable[[Callable], Callable]
LocaleType: TypeAlias = str | tuple[str | None, str | None]

DecoratorClassT = TypeVar('DecoratorClassT', covariant=True)
JobConfigT = TypeVar('JobConfigT', covariant=True)
JobT = TypeVar('JobT', covariant=True)
JobTemplateT = TypeVar('JobTemplateT', covariant=True)
TaskTemplateT = TypeVar('TaskTemplateT')
TaskTemplateContraT = TypeVar('TaskTemplateContraT', contravariant=True)
TaskTemplateCovT = TypeVar('TaskTemplateCovT', covariant=True)

# TODO: While waiting for https://github.com/python/mypy/issues/9773
TypeForm: TypeAlias = (
    type | UnionType | None | _UnionGenericAlias | _AnnotatedAlias | GenericAlias
    | _LiteralGenericAlias | _SpecialForm | ForwardRef | TypeVar)



# generated by datamodel-codegen:
#   filename:  process_schema.json
#   timestamp: 2024-02-07T08:09:26+00:00

from datetime import date, datetime
from enum import Enum
from typing import List, Optional, Union

from pydantic import BaseModel, constr, Extra, Field, validator

from omnipy.data.model import Model

from . import (comment_schema,
               data_schema,
               material_schema,
               process_parameter_value_schema,
               protocol_schema,
               sample_schema,
               source_schema)
from .validators import date_to_iso_format


class FieldType(Enum):
    Process = 'Process'


class IsaProcessOrProtocolApplicationSchema(BaseModel):
    class Config:
        extra = Extra.forbid
        use_enum_values = True

    field_id: Optional[str] = Field(None, alias='@id')
    field_context: Optional[str] = Field(None, alias='@context')
    field_type: Optional[FieldType] = Field(None, alias='@type')
    name: Optional[str] = None
    executesProtocol: Optional[protocol_schema.IsaProtocolModel] = None
    parameterValues: Optional[List[
        process_parameter_value_schema.IsaProcessParameterValueModel]] = None
    performer: Optional[str] = None
    date: Optional[Union[datetime, date, constr(max_length=0)]] = None
    previousProcess: Optional['IsaProcessOrProtocolApplicationModel'] = None
    nextProcess: Optional['IsaProcessOrProtocolApplicationModel'] = None
    inputs: Optional[List[Union[
        source_schema.IsaSourceModel,
        sample_schema.IsaSampleModel,
        data_schema.IsaDataModel,
        material_schema.IsaMaterialModel,
    ]]] = None
    outputs: Optional[List[Union[
        sample_schema.IsaSampleModel,
        data_schema.IsaDataModel,
        material_schema.IsaMaterialModel,
    ]]] = None
    comments: Optional[List[comment_schema.IsaCommentModel]] = None

    _date_to_iso_format = validator('date', allow_reuse=True)(date_to_iso_format)


class IsaProcessOrProtocolApplicationModel(Model[IsaProcessOrProtocolApplicationSchema]):
    ...


IsaProcessOrProtocolApplicationSchema.update_forward_refs()



from typing import Callable, ContextManager, Protocol, TypeVar

from omnipy.api.protocols.private.util import HasContents, IsSnapshotHolder
from omnipy.api.protocols.public.config import IsDataConfig

_ObjT = TypeVar('_ObjT', bound=HasContents)
_ContentsT = TypeVar('_ContentsT', bound=object)


class IsDataClassCreator(Protocol[_ObjT, _ContentsT]):
    """"""
    @property
    def config(self) -> IsDataConfig:
        ...

    def set_config(self, config: IsDataConfig) -> None:
        ...

    @property
    def snapshot_holder(self) -> IsSnapshotHolder[_ObjT, _ContentsT]:
        ...

    def deepcopy_context(
        self,
        top_level_entry_func: Callable[[], None],
        top_level_exit_func: Callable[[], None],
    ) -> ContextManager[int]:
        ...


class IsDataClassBase(Protocol):
    """"""
    @property
    def _data_class_creator(self) -> IsDataClassCreator:
        ...

    @property
    def config(self) -> IsDataConfig:
        ...



from datetime import datetime
from typing import Protocol

from omnipy.api.protocols.private.compute.job import (IsFuncArgJob,
                                                      IsFuncArgJobTemplate,
                                                      IsTaskTemplateArgsJob,
                                                      IsTaskTemplateArgsJobTemplate)
from omnipy.api.protocols.private.compute.mixins import IsNestedContext


class IsTaskTemplate(IsFuncArgJobTemplate['IsTaskTemplate', 'IsTask'], Protocol):
    """
    Loosely coupled type replacement for the :py:class:`~omnipy.compute.task.TaskTemplate` class
    """
    ...


class IsTask(IsFuncArgJob[IsTaskTemplate], Protocol):
    """"""
    ...


class IsFlowTemplate(Protocol):
    """"""
    ...


class IsFlow(Protocol):
    """"""
    @property
    def flow_context(self) -> IsNestedContext:
        ...

    @property
    def time_of_last_run(self) -> datetime | None:
        ...


class IsLinearFlowTemplate(IsTaskTemplateArgsJobTemplate[IsTaskTemplate,
                                                         'IsLinearFlowTemplate',
                                                         'IsLinearFlow'],
                           IsFlowTemplate,
                           Protocol):
    """"""
    ...


class IsLinearFlow(IsTaskTemplateArgsJob[IsTaskTemplate, IsLinearFlowTemplate], IsFlow, Protocol):
    """"""
    ...


class IsDagFlowTemplate(IsTaskTemplateArgsJobTemplate[IsTaskTemplate,
                                                      'IsDagFlowTemplate',
                                                      'IsDagFlow'],
                        IsFlowTemplate,
                        Protocol):
    """"""
    ...


class IsDagFlow(IsTaskTemplateArgsJob[IsTaskTemplate, IsDagFlowTemplate], IsFlow, Protocol):
    """"""
    ...


class IsFuncFlowTemplate(IsFuncArgJobTemplate['IsFuncFlowTemplate', 'IsFuncFlow'],
                         IsFlowTemplate,
                         Protocol):
    """"""
    ...


class IsFuncFlow(IsFuncArgJob[IsFuncFlowTemplate], Protocol):
    """"""
    ...



# generated by datamodel-codegen:
#   filename:  study_group.json
#   timestamp: 2024-02-07T08:09:26+00:00

from enum import Enum
from typing import List, Optional

from pydantic import BaseModel, Extra, Field

from omnipy.data.model import Model

from . import comment_schema, factor_value_schema, sample_schema


class FieldType(Enum):
    Study_Group = 'Study Group'


class IsaStudyGroupSchema(BaseModel):
    class Config:
        extra = Extra.forbid
        use_enum_values = True

    field_id: Optional[str] = Field(None, alias='@id')
    field_context: Optional[str] = Field(None, alias='@context')
    field_type: Optional[FieldType] = Field(None, alias='@type')
    name: Optional[str] = None
    factor_levels: Optional[List[factor_value_schema.IsaFactorValueModel]] = None
    study_group_size: Optional[int] = None
    members: Optional[List[sample_schema.IsaSampleModel]] = None
    comments: Optional[List[comment_schema.IsaCommentModel]] = None


class IsaStudyGroupModel(Model[IsaStudyGroupSchema]):
    ...



from typing import Protocol, runtime_checkable, Type

from omnipy.api.protocols.private.log import IsRunStateRegistry
from omnipy.api.protocols.public.config import IsEngineConfig


@runtime_checkable
class IsEngine(Protocol):
    """"""
    def __init__(self) -> None:
        ...

    @classmethod
    def get_config_cls(cls) -> Type[IsEngineConfig]:
        ...

    def set_config(self, config: IsEngineConfig) -> None:
        ...

    def set_registry(self, registry: IsRunStateRegistry | None) -> None:
        ...



from collections import defaultdict
from typing import Protocol, TextIO

from omnipy.api.enums import (BackoffStrategy,
                              ConfigOutputStorageProtocolOptions,
                              ConfigPersistOutputsOptions,
                              ConfigRestoreOutputsOptions)
from omnipy.api.typedefs import LocaleType


class IsEngineConfig(Protocol):
    """"""
    ...


class IsLocalRunnerConfig(IsEngineConfig, Protocol):
    """"""
    ...


class IsPrefectEngineConfig(IsEngineConfig, Protocol):
    """"""
    use_cached_results: bool = False


class IsJobConfig(Protocol):
    """"""
    output_storage: 'IsOutputStorageConfig'


class IsDataConfig(Protocol):
    """"""
    interactive_mode: bool
    dynamically_convert_elements_to_models: bool
    terminal_size_columns: int
    terminal_size_lines: int
    http_defaults: 'IsHttpConfig'
    http_config_for_host: defaultdict[str, 'IsHttpConfig']


class IsHttpConfig(Protocol):
    """"""
    requests_per_time_period: float
    time_period_in_secs: float
    retry_http_statuses: tuple[int, ...]
    retry_attempts: int
    retry_backoff_strategy: BackoffStrategy


class IsRootLogConfig(Protocol):
    """"""
    log_format_str: str
    locale: LocaleType
    log_to_stdout: bool
    log_to_stderr: bool
    log_to_file: bool
    stdout: TextIO
    stderr: TextIO
    stdout_log_min_level: int
    stderr_log_min_level: int
    file_log_min_level: int
    file_log_path: str


class IsOutputStorageConfigBase(Protocol):
    persist_data_dir_path: str


class IsLocalOutputStorageConfig(IsOutputStorageConfigBase, Protocol):
    ...


class IsS3OutputStorageConfig(IsOutputStorageConfigBase, Protocol):
    endpoint_url: str
    access_key: str
    secret_key: str
    bucket_name: str


class IsOutputStorageConfig(Protocol):
    persist_outputs: ConfigPersistOutputsOptions
    restore_outputs: ConfigRestoreOutputsOptions
    protocol: ConfigOutputStorageProtocolOptions
    local: IsLocalOutputStorageConfig
    s3: IsS3OutputStorageConfig



from contextlib import AbstractContextManager, contextmanager
from copy import deepcopy
from typing import Callable, Iterator, ParamSpec

from typing_extensions import TypeVar

# TODO: Consider refactoring as many as possible of the context managers (AbstractContextManager
#       subclasses) to @contextmanager-decorated methods

_SetupP = ParamSpec('_SetupP')
_TeardownP = ParamSpec('_TeardownP')
_ValT = TypeVar('_ValT')


@contextmanager
def setup_and_teardown_callback_context(
    *,
    setup_func: Callable[_SetupP, _ValT] | None = None,
    setup_func_args: _SetupP.args = (),
    setup_func_kwargs: _SetupP.kwargs = {},
    exception_func: Callable[_TeardownP, None] | None = None,
    exception_func_args: _TeardownP.args = (),
    exception_func_kwargs: _TeardownP.kwargs = {},
    teardown_func: Callable[_TeardownP, None] | None = None,
    teardown_func_args: _TeardownP.args = (),
    teardown_func_kwargs: _TeardownP.kwargs = {},
) -> Iterator[_ValT | None]:
    setup_val: _ValT | None = None
    if setup_func is not None:
        setup_val = setup_func(*setup_func_args, **setup_func_kwargs)
    try:
        yield setup_val
    except Exception:
        if exception_func is not None:
            exception_func(*exception_func_args, **exception_func_kwargs)
        raise
    finally:
        if teardown_func is not None:
            teardown_func(*teardown_func_args, **teardown_func_kwargs)


class LastErrorHolder(AbstractContextManager):
    def __init__(self):
        self._last_error = None

    def __enter__(self):
        ...

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val is not None:
            self._last_error = exc_val
        return True

    def raise_derived(self, exc: Exception):
        if self._last_error is not None:
            raise exc from self._last_error
        else:
            raise exc


@contextmanager
def hold_and_reset_prev_attrib_value(
    obj: object,
    attr_name: str,
    copy_attr: bool = False,
) -> Iterator[None]:
    attr_value = getattr(obj, attr_name)
    prev_value = deepcopy(attr_value) if copy_attr else attr_value

    try:
        yield
    finally:
        setattr(obj, attr_name, prev_value)


@contextmanager
def nothing(*args, **kwds) -> Iterator[None]:
    yield None


class PrintExceptionContext(AbstractContextManager):
    def __enter__(self):
        ...

    def __exit__(self, exc_type, exc_val, exc_tb):
        print(f'{exc_type.__name__}: {str(exc_val).splitlines()[0]}', end='')
        return True


print_exception = PrintExceptionContext()



import ast
from inspect import formatannotation, isgeneratorfunction, Signature
import os
from types import ModuleType
from typing import Any, get_type_hints

from docstring_parser import DocstringParam, DocstringReturns
from pdocs.doc import Doc, External, Function, Module

from omnipy.util.helpers import recursive_module_import

IGNORED = None
IGNORE_PARAMS = ['cls', 'self']


def filter_internal_external(cls: type, members: list[Doc]):
    internal_members = []
    external_members = []

    for member in members:
        try:
            if is_member(cls, member.name) or is_internally_inherited(cls, member.name, ['omnipy']):
                internal_members.append(member)
            else:
                external_members.append(member)
        except AttributeError as e:
            print(e)

    return internal_members, external_members


def is_member(cls: type, member_name: str):
    return member_name in vars(cls)


def is_internally_inherited(cls: type,
                            member_name: str,
                            internal_packages: list[str],
                            outer: bool = True):
    if outer and member_name not in dir(cls):
        raise AttributeError(f'"{member_name}" is not a member of class "{cls.__name__}"')

    if is_member(cls, member_name):
        if outer:
            return False
        else:
            is_internal = any(cls.__module__.startswith(pkg) for pkg in internal_packages)
            return is_internal
    else:
        for base_cls in cls.__bases__:
            if is_internally_inherited(base_cls, member_name, internal_packages, outer=False):
                return True
        return False


def merge_signature_with_docstring(func: Function,
                                   signature: Signature,
                                   ds_params: list[DocstringParam],
                                   ds_returns: DocstringReturns):
    ds_params_map = {ds_param.arg_name: ds_param for ds_param in ds_params}
    params = []

    for name, param in signature.parameters.items():
        if name in IGNORE_PARAMS:
            continue

        description = ''
        if name in ds_params_map:
            description = ds_params_map[name].description

        type_name = get_type_name_from_annotation(func.module.module, param.annotation, param.empty)

        default = param.default if param.default is not param.empty else ''

        params.append(
            DocstringParam(
                args=[],
                description=description,
                arg_name=name,
                type_name=type_name,
                is_optional=IGNORED,
                default=default))

    description = ds_returns.description if ds_returns else ''

    if type_name := get_type_name_from_annotation(func.module.module,
                                                  signature.return_annotation,
                                                  signature.empty):
        returns = DocstringReturns(
            args=[],
            description=description,
            type_name=type_name,
            is_generator=isgeneratorfunction(func.func),
            return_name=IGNORED)
    else:
        returns = None

    return params, returns


def get_type_name_from_annotation(module: ModuleType, annotation, empty_obj):
    if annotation is not empty_obj:
        # (f'Before: {repr(annotation)}')
        type_name = convert_to_qual_name_type_hint_str(module, annotation)
        # print(f'After: {repr(type_name)}')
    else:
        type_name = ''

    return type_name


def convert_to_qual_name_type_hint_str(module: ModuleType, type_hint: Any) -> str:
    def fixed_get_type_hints(obj: Any) -> str:
        """
        Workaround from https://stackoverflow.com/a/66845686, due to limitations in get_type_hints()
        per Python 3.10
        """
        try:

            class X:
                x: obj

            return get_type_hints(X, globalns=recursive_module_import(module))['x']
        except NameError as e:
            print(e)
            return obj

    return formatannotation(fixed_get_type_hints(type_hint))


def format_type_as_markdown(module, type_name: str):
    cleaned_type_name = cleanup_type_hint_str_before_parsing(type_name)
    for basic_type_name in set(parse_type_hint(cleaned_type_name)):
        refname, url = lookup(module, basic_type_name)
        if refname:
            type_name = type_name.replace(basic_type_name, f'[{refname}]({url})')
    return type_name


_TYPE_HINTS_REPLACE_MAP = {
    '~': '',
    '+': '',
    '-': '',
}


def cleanup_type_hint_str_before_parsing(type_hint):
    for from_str, to_str in _TYPE_HINTS_REPLACE_MAP.items():
        type_hint = type_hint.replace(from_str, to_str)
    return type_hint


def parse_type_hint(type_hint_string):
    """
    Parses a Python type hint string and returns the individual types as a list of strings.

    Generated by ChatGPT May 21, 2023, in a guided session. This is the 22nd attempt.

    Args:
        type_hint_string (str): The type hint string to parse.

    Returns:
        list: The individual types parsed from the type hint string.

    Example:
        type_hint = "Union[list[str], dict[str, int], tuple[int, str], Optional[int]]"
        parsed_types = parse_type_hint(type_hint)
        print(parsed_types)
        # Output: ['Union', 'list', 'str', 'dict', 'str', 'int', 'tuple', 'int',
                   'str', 'Optional', 'int']
    """
    # print(f'Parsing: {repr(type_hint_string)}')
    source = f'def f() -> {type_hint_string}: pass'
    tree = ast.parse(source)

    type_strings = []

    def process_node(node, qual_names=None):
        """
        Recursively processes the AST nodes and extracts the type names.

        Args:
            node (ast.AST): The AST node to process.
            qual_names (list): A list of qualified names encountered during processing.

        Returns:
            None
        """
        qual_names = qual_names or []

        if isinstance(node, ast.Name):
            type_strings.append('.'.join(reversed(qual_names + [node.id])))

        elif isinstance(node, ast.Subscript):
            process_node(node.value, qual_names)
            process_node(node.slice, qual_names)

        elif isinstance(node, ast.Attribute):
            process_node(node.value, qual_names + [node.attr])

        elif isinstance(node, ast.Tuple):
            for elt in node.elts:
                process_node(elt, qual_names)

    returns_annotation = tree.body[0].returns
    process_node(returns_annotation)

    return type_strings


def lookup(module, refname):
    """
    Adapted from https://github.com/timothycrosley/pdocs/blob/master/pdocs/html_helpers.py
    """

    if '.' not in refname:
        refname = f'{module.name}.{refname}'

    d = module.find_ident(refname)

    if isinstance(d, External):
        return None, None
    if isinstance(d, Module):
        return d.refname, module_url(module, d)

    return d.name, f'{module_url(module, d.module)}#{d.name.lower()}'


def module_url(parent, m):
    """
    Adapted from https://github.com/timothycrosley/pdocs/blob/master/pdocs/html_helpers.py
    """

    relpath = os.path.relpath(m.name.replace('.', '/'), parent.name.replace('.', '/'))
    if relpath == '.':
        return ''
    else:
        return relpath + '/'



from omnipy.compute.flow import FuncFlowTemplate
from omnipy.compute.typing import mypy_fix_func_flow_template

from .datasets import (JsonDictOfDictsDataset,
                       JsonDictOfListsOfDictsDataset,
                       JsonListOfDictsDataset,
                       JsonListOfDictsOfScalarsDataset)
from .tasks import (_flatten_outer_level_of_all_data_files,
                    DEFAULT_KEY,
                    ID_KEY,
                    REF_KEY,
                    transpose_dicts_2_lists)


@mypy_fix_func_flow_template
@FuncFlowTemplate(fixed_params=dict(id_key=ID_KEY, ref_key=REF_KEY, default_key=DEFAULT_KEY))
def flatten_nested_json(
    dataset: JsonListOfDictsDataset,
    id_key: str,
    ref_key: str,
    default_key: str,
) -> JsonListOfDictsOfScalarsDataset:
    all_data_files_of_scalar_records_ds = JsonListOfDictsOfScalarsDataset()
    data_files_of_any_ds = dataset

    while len(data_files_of_any_ds) > 0:
        data_files_of_scalar_records_ds, data_files_of_any_ds = \
            _flatten_outer_level_of_all_data_files(
                data_files_of_any_ds, id_key, ref_key, default_key)

        all_data_files_of_scalar_records_ds |= data_files_of_scalar_records_ds

    return all_data_files_of_scalar_records_ds


@mypy_fix_func_flow_template
@FuncFlowTemplate()
def transpose_dict_of_dicts_2_list_of_dicts(
    dataset: JsonDictOfDictsDataset,
    id_key: str = ID_KEY,
) -> JsonListOfDictsDataset:
    output_dataset = JsonListOfDictsDataset()
    output_dataset |= transpose_dicts_2_lists(dataset, id_key=id_key)
    return output_dataset


@mypy_fix_func_flow_template
@FuncFlowTemplate()
def transpose_dicts_of_lists_of_dicts_2_lists_of_dicts(
    dataset: JsonDictOfListsOfDictsDataset,
    id_key: str = ID_KEY,
) -> JsonListOfDictsDataset:
    output_dataset = JsonListOfDictsDataset()
    output_dataset |= transpose_dicts_2_lists(dataset, id_key=id_key)
    return output_dataset



import logging
from logging.handlers import RotatingFileHandler
from typing import Protocol

from omnipy.api.enums import EngineChoice
from omnipy.api.protocols.private.compute.job_creator import IsJobConfigHolder
from omnipy.api.protocols.private.data import IsDataClassCreator
from omnipy.api.protocols.private.engine import IsEngine
from omnipy.api.protocols.private.log import IsRunStateRegistry
from omnipy.api.protocols.public.config import (IsDataConfig,
                                                IsJobConfig,
                                                IsLocalRunnerConfig,
                                                IsPrefectEngineConfig,
                                                IsRootLogConfig)
from omnipy.api.protocols.public.data import IsSerializerRegistry


class IsRootLogObjects(Protocol):
    """"""
    formatter: logging.Formatter | None = None
    stdout_handler: logging.StreamHandler | None = None
    stderr_handler: logging.StreamHandler | None = None
    file_handler: RotatingFileHandler | None = None

    def set_config(self, config: IsRootLogConfig) -> None:
        ...


class IsRuntimeConfig(Protocol):
    """"""
    job: IsJobConfig
    data: IsDataConfig
    engine: EngineChoice
    local: IsLocalRunnerConfig
    prefect: IsPrefectEngineConfig
    root_log: IsRootLogConfig

    def __init__(
            self,
            job: IsJobConfig | None = None,  # noqa
            data: IsDataConfig | None = None,  # noqa
            engine: EngineChoice = EngineChoice.LOCAL,  # noqa
            local: IsLocalRunnerConfig | None = None,  # noqa
            prefect: IsPrefectEngineConfig | None = None,  # noqa
            root_log: IsRootLogConfig | None = None,  # noqa
            *args: object,
            **kwargs: object) -> None:
        ...

    def reset_to_defaults(self):
        ...


class IsRuntimeObjects(Protocol):
    """"""

    job_creator: IsJobConfigHolder
    data_class_creator: IsDataClassCreator
    local: IsEngine
    prefect: IsEngine
    registry: IsRunStateRegistry
    serializers: IsSerializerRegistry
    root_log: IsRootLogObjects

    def __init__(
            self,
            job_creator: IsJobConfigHolder | None = None,  # noqa
            data_class_creator: IsDataClassCreator | None = None,  # noqa
            local: IsEngine | None = None,  # noqa
            prefect: IsEngine | None = None,  # noqa
            registry: IsRunStateRegistry | None = None,  # noqa
            serializers: IsSerializerRegistry | None = None,  # noqa
            root_log: IsRootLogObjects | None = None,  # noqa
            *args: object,
            **kwargs: object) -> None:
        ...


class IsRuntime(Protocol):
    """"""
    config: IsRuntimeConfig
    objects: IsRuntimeObjects

    def __init__(
            self,
            config: IsRuntimeConfig | None = None,  # noqa
            objects: IsRuntimeObjects | None = None,  # noqa
            *args: object,
            **kwargs: object) -> None:
        ...

    def reset_subscriptions(self):
        ...



from typing import cast, Type

from omnipy.api.protocols.private.compute.job import (IsFuncArgJobTemplateCallable,
                                                      IsJob,
                                                      IsJobTemplate,
                                                      IsTaskTemplateArgsJobTemplateCallable)
from omnipy.api.protocols.private.engine import IsEngine
from omnipy.api.protocols.public.compute import (IsDagFlow,
                                                 IsDagFlowTemplate,
                                                 IsFuncFlow,
                                                 IsFuncFlowTemplate,
                                                 IsLinearFlow,
                                                 IsLinearFlowTemplate,
                                                 IsTaskTemplate)
from omnipy.api.protocols.public.engine import (IsDagFlowRunnerEngine,
                                                IsFuncFlowRunnerEngine,
                                                IsLinearFlowRunnerEngine)
from omnipy.compute.func_job import FuncArgJobBase
from omnipy.compute.job import JobMixin, JobTemplateMixin
from omnipy.compute.mixins.flow_context import FlowContextJobMixin
from omnipy.compute.tasklist_job import TaskTemplateArgsJobBase
from omnipy.util.callable_decorator import callable_decorator_cls


class FlowBase:
    ...


def linear_flow_template_callable_decorator_cls(
    cls: Type['LinearFlowTemplate']
) -> IsTaskTemplateArgsJobTemplateCallable[IsTaskTemplate, IsLinearFlowTemplate]:
    return cast(IsTaskTemplateArgsJobTemplateCallable[IsTaskTemplate, IsLinearFlowTemplate],
                callable_decorator_cls(cls))


@linear_flow_template_callable_decorator_cls
class LinearFlowTemplate(JobTemplateMixin, FlowBase, TaskTemplateArgsJobBase):
    @classmethod
    def _get_job_subcls_for_apply(cls) -> Type[IsJob]:
        return cast(Type[IsLinearFlow], LinearFlow)


class LinearFlow(JobMixin, FlowBase, TaskTemplateArgsJobBase):
    def _apply_engine_decorator(self, engine: IsEngine) -> None:
        if self.engine:
            engine = cast(IsLinearFlowRunnerEngine, self.engine)
            self_with_mixins = cast(IsLinearFlow, self)
            engine.apply_linear_flow_decorator(self_with_mixins, self._accept_call_func_decorator)

    @classmethod
    def _get_job_template_subcls_for_revise(cls) -> Type[IsJobTemplate]:
        return cast(Type[IsLinearFlowTemplate], LinearFlowTemplate)


def dag_flow_template_callable_decorator_cls(
    cls: Type['DagFlowTemplate']
) -> IsTaskTemplateArgsJobTemplateCallable[IsTaskTemplate, IsDagFlowTemplate]:
    return cast(IsTaskTemplateArgsJobTemplateCallable[IsTaskTemplate, IsDagFlowTemplate],
                callable_decorator_cls(cls))


@dag_flow_template_callable_decorator_cls
class DagFlowTemplate(JobTemplateMixin, FlowBase, TaskTemplateArgsJobBase):
    @classmethod
    def _get_job_subcls_for_apply(cls) -> Type[IsJob]:
        return cast(Type[IsDagFlow], DagFlow)


class DagFlow(JobMixin, FlowBase, TaskTemplateArgsJobBase):
    def _apply_engine_decorator(self, engine: IsEngine) -> None:
        if self.engine:
            engine = cast(IsDagFlowRunnerEngine, self.engine)
            self_with_mixins = cast(IsDagFlow, self)
            engine.apply_dag_flow_decorator(self_with_mixins, self._accept_call_func_decorator)

    @classmethod
    def _get_job_template_subcls_for_revise(cls) -> Type[IsJobTemplate]:
        return cast(Type[IsDagFlowTemplate], DagFlowTemplate)


def func_flow_template_callable_decorator_cls(
        cls: Type['FuncFlowTemplate']) -> IsFuncArgJobTemplateCallable[IsFuncFlowTemplate]:
    return cast(IsFuncArgJobTemplateCallable[IsFuncFlowTemplate], callable_decorator_cls(cls))


@func_flow_template_callable_decorator_cls
class FuncFlowTemplate(JobTemplateMixin, FlowBase, FuncArgJobBase):
    @classmethod
    def _get_job_subcls_for_apply(cls) -> Type[IsJob]:
        return cast(Type[IsFuncFlow], FuncFlow)


class FuncFlow(JobMixin, FlowBase, FuncArgJobBase):
    def _apply_engine_decorator(self, engine: IsEngine) -> None:
        if self.engine:
            engine = cast(IsFuncFlowRunnerEngine, self.engine)
            self_with_mixins = cast(IsFuncFlow, self)
            engine.apply_func_flow_decorator(self_with_mixins, self._accept_call_func_decorator)

    @classmethod
    def _get_job_template_subcls_for_revise(cls) -> Type[IsJobTemplate]:
        return cast(Type[IsFuncFlowTemplate], FuncFlowTemplate)


LinearFlow.accept_mixin(FlowContextJobMixin)
DagFlow.accept_mixin(FlowContextJobMixin)
FuncFlow.accept_mixin(FlowContextJobMixin)

# TODO: Recursive replace - *args: Any -> *args: object, *kwargs: Any -> *kwargs: object



import pandas as pd

from .models import PandasModel


def extract_common_colnames(df_1: PandasModel | pd.DataFrame, df_2: PandasModel | pd.DataFrame):
    common_colnames_set = set(df_1.columns) & set(df_2.columns)
    common_colnames = tuple(col for col in df_1.columns if col in common_colnames_set)
    return common_colnames



from typing import Generic

from typing_extensions import TypeVar

from omnipy.data.dataset import Dataset
from omnipy.data.param import bind_adjust_dataset_func

from .models import (BytesModel,
                     JoinColumnsToLinesModel,
                     JoinItemsModel,
                     JoinLinesModel,
                     JoinSubitemsToItemsModel,
                     SplitItemsToSubitemsModel,
                     SplitLinesToColumnsModel,
                     SplitToItemsModel,
                     SplitToLinesModel,
                     StrModel)

BytesModelT = TypeVar('BytesModelT', default=BytesModel)
StrModelT = TypeVar('StrModelT', default=StrModel)
SplitToItemsModelT = TypeVar('SplitToItemsModelT', default=SplitToItemsModel)
JoinItemsModelT = TypeVar('JoinItemsModelT', default=JoinItemsModel)
SplitItemsToSubitemsModelT = TypeVar(
    'SplitItemsToSubitemsModelT', default=SplitItemsToSubitemsModel)
JoinSubitemsToItemsModelT = TypeVar('JoinSubitemsToItemsModelT', default=JoinSubitemsToItemsModel)


class _BytesDataset(Dataset[BytesModelT], Generic[BytesModelT]):
    ...


class BytesDataset(_BytesDataset[BytesModel]):
    adjust = bind_adjust_dataset_func(
        _BytesDataset[BytesModel].clone_dataset_cls,
        BytesModel,
        BytesModel.Params,
    )


class _StrDataset(Dataset[StrModelT], Generic[StrModelT]):
    ...


class StrDataset(_StrDataset[StrModel]):
    adjust = bind_adjust_dataset_func(
        _StrDataset[StrModel].clone_dataset_cls,
        StrModel,
        StrModel.Params,
    )


class _SplitToItemsDataset(Dataset[SplitToItemsModelT], Generic[SplitToItemsModelT]):
    ...


class SplitToItemsDataset(_SplitToItemsDataset[SplitToItemsModel]):
    adjust = bind_adjust_dataset_func(
        _SplitToItemsDataset[SplitToItemsModel].clone_dataset_cls,
        SplitToItemsModel,
        SplitToItemsModel.Params,
    )


class SplitToLinesDataset(_SplitToItemsDataset[SplitToLinesModel]):
    adjust = bind_adjust_dataset_func(
        _SplitToItemsDataset[SplitToLinesModel].clone_dataset_cls,
        SplitToLinesModel,
        SplitToLinesModel.Params,
    )


class _JoinItemsDataset(Dataset[JoinItemsModelT], Generic[JoinItemsModelT]):
    ...


class JoinItemsDataset(_JoinItemsDataset[JoinItemsModel]):
    adjust = bind_adjust_dataset_func(
        _JoinItemsDataset[JoinItemsModel].clone_dataset_cls,
        JoinItemsModel,
        JoinItemsModel.Params,
    )


class JoinLinesDataset(_JoinItemsDataset[JoinLinesModel]):
    adjust = bind_adjust_dataset_func(
        _JoinItemsDataset[JoinLinesModel].clone_dataset_cls,
        JoinLinesModel,
        JoinLinesModel.Params,
    )


class _SplitItemsToSubitemsDataset(Dataset[SplitItemsToSubitemsModelT],
                                   Generic[SplitItemsToSubitemsModelT]):
    ...


class SplitLinesToColumnsDataset(_SplitItemsToSubitemsDataset[SplitLinesToColumnsModel]):
    adjust = bind_adjust_dataset_func(
        _SplitItemsToSubitemsDataset[SplitLinesToColumnsModel].clone_dataset_cls,
        SplitLinesToColumnsModel,
        SplitLinesToColumnsModel.Params,
    )


class _JoinSubitemsToItemsDataset(Dataset[JoinSubitemsToItemsModelT],
                                  Generic[JoinSubitemsToItemsModelT]):
    ...


class JoinColumnsToLinesDataset(_JoinSubitemsToItemsDataset[JoinColumnsToLinesModel]):
    adjust = bind_adjust_dataset_func(
        _JoinSubitemsToItemsDataset[JoinColumnsToLinesModel].clone_dataset_cls,
        JoinColumnsToLinesModel,
        JoinColumnsToLinesModel.Params,
    )



from datetime import datetime
from inspect import BoundArguments
from types import MappingProxyType
from typing import Any, Callable, Mapping, Protocol, Type

from omnipy.api.enums import (OutputStorageProtocolOptions,
                              PersistOutputsOptions,
                              RestoreOutputsOptions)
from omnipy.api.protocols.private.compute.job_creator import IsJobCreator
from omnipy.api.protocols.private.compute.mixins import IsUniquelyNamedJob
from omnipy.api.protocols.private.engine import IsEngine
from omnipy.api.protocols.private.log import CanLog
from omnipy.api.protocols.public.config import IsJobConfig
from omnipy.api.protocols.public.data import IsDataset
from omnipy.api.typedefs import (GeneralDecorator,
                                 JobT,
                                 JobTemplateT,
                                 TaskTemplateContraT,
                                 TaskTemplateCovT,
                                 TaskTemplateT)


class IsJobBase(CanLog, IsUniquelyNamedJob, Protocol):
    """"""
    @property
    def _job_creator(self) -> IsJobCreator:
        ...

    @property
    def config(self) -> IsJobConfig | None:
        ...

    @property
    def engine(self) -> IsEngine | None:
        ...

    @property
    def in_flow_context(self) -> bool:
        ...

    def __eq__(self, other: object):
        ...

    @classmethod
    def _create_job_template(cls, *args: object, **kwargs: object) -> 'IsJobTemplate':
        ...

    @classmethod
    def _create_job(cls, *args: object, **kwargs: object) -> 'IsJob':
        ...

    def _apply(self) -> 'IsJob':
        ...

    def _refine(self, *args: Any, update: bool = True, **kwargs: object) -> 'IsJobTemplate':
        ...

    def _revise(self) -> 'IsJobTemplate':
        ...

    def _call_job_template(self, *args: object, **kwargs: object) -> object:
        ...

    def _call_job(self, *args: object, **kwargs: object) -> object:
        ...


class IsJob(IsJobBase, Protocol):
    """"""
    @property
    def time_of_cur_toplevel_flow_run(self) -> datetime | None:
        ...

    @classmethod
    def create_job(cls, *args: object, **kwargs: object) -> 'IsJob':
        ...

    def __call__(self, *args: object, **kwargs: object) -> object:
        ...

    def _apply_engine_decorator(self, engine: IsEngine) -> None:
        ...


class IsJobTemplate(IsJobBase, Protocol):
    """"""
    @classmethod
    def create_job_template(cls, *args: object, **kwargs: object) -> 'IsJobTemplate':
        ...

    def run(self, *args: object, **kwargs: object) -> object:
        ...


class IsFuncArgJobBase(IsJob, Protocol):
    """"""
    @property
    def param_signatures(self) -> MappingProxyType:
        ...

    @property
    def return_type(self) -> Type[object]:
        ...

    @property
    def iterate_over_data_files(self) -> bool:
        ...

    @property
    def output_dataset_param(self) -> str | None:
        ...

    @property
    def output_dataset_cls(self) -> type[IsDataset] | None:
        ...

    @property
    def auto_async(self) -> bool:
        ...

    @property
    def persist_outputs(self) -> PersistOutputsOptions | None:
        ...

    @property
    def restore_outputs(self) -> RestoreOutputsOptions | None:
        ...

    @property
    def output_storage_protocol(self) -> OutputStorageProtocolOptions | None:
        ...

    @property
    def will_persist_outputs(self) -> PersistOutputsOptions:
        ...

    @property
    def will_restore_outputs(self) -> RestoreOutputsOptions:
        ...

    @property
    def output_storage_protocol_to_use(self) -> OutputStorageProtocolOptions:
        ...

    @property
    def result_key(self) -> str | None:
        ...

    @property
    def fixed_params(self) -> MappingProxyType[str, object]:
        ...

    @property
    def param_key_map(self) -> MappingProxyType[str, str]:
        ...

    def has_coroutine_func(self) -> bool:
        ...

    def get_bound_args(self, *args: object, **kwargs: object) -> BoundArguments:
        ...


class IsPlainFuncArgJobBase(Protocol):
    """"""
    _job_func: Callable

    def _accept_call_func_decorator(self, call_func_decorator: GeneralDecorator) -> None:
        ...


class IsFuncArgJob(IsFuncArgJobBase, Protocol[JobT]):
    """"""
    def revise(self) -> JobT:
        ...


class IsFuncArgJobTemplateCallable(Protocol[
        JobTemplateT,
]):
    """"""
    def __call__(
        self,
        name: str | None = None,
        iterate_over_data_files: bool = False,
        output_dataset_param: str | None = None,
        output_dataset_cls: type[IsDataset] | None = None,
        auto_async: bool = True,
        persist_outputs: PersistOutputsOptions | None = None,
        restore_outputs: RestoreOutputsOptions | None = None,
        result_key: str | None = None,
        fixed_params: Mapping[str, object] | None = None,
        param_key_map: Mapping[str, str] | None = None,
        **kwargs: object,
    ) -> Callable[[Callable], JobTemplateT]:
        ...


class IsFuncArgJobTemplate(IsJobTemplate, IsFuncArgJobBase, Protocol[JobTemplateT, JobT]):
    """"""
    def refine(self,
               *args: Any,
               update: bool = True,
               name: str | None = None,
               iterate_over_data_files: bool = False,
               output_dataset_param: str | None = None,
               output_dataset_cls: type[IsDataset] | None = None,
               auto_async: bool = True,
               persist_outputs: PersistOutputsOptions | None = None,
               restore_outputs: RestoreOutputsOptions | None = None,
               result_key: str | None = None,
               fixed_params: Mapping[str, object] | None = None,
               param_key_map: Mapping[str, str] | None = None,
               **kwargs: object) -> JobTemplateT:
        ...

    def apply(self) -> JobT:
        ...


class IsTaskTemplateArgsJobBase(IsFuncArgJobBase, Protocol[TaskTemplateCovT]):
    """"""
    @property
    def task_templates(self) -> tuple[TaskTemplateCovT, ...]:
        ...


class IsTaskTemplateArgsJob(IsTaskTemplateArgsJobBase[TaskTemplateCovT],
                            IsFuncArgJob[JobT],
                            Protocol[TaskTemplateCovT, JobT]):
    """"""


class IsTaskTemplateArgsJobTemplateCallable(Protocol[TaskTemplateContraT, JobTemplateT]):
    """"""
    def __call__(
        self,
        *task_templates: TaskTemplateContraT,
        name: str | None = None,
        iterate_over_data_files: bool = False,
        output_dataset_param: str | None = None,
        output_dataset_cls: type[IsDataset] | None = None,
        auto_async: bool = True,
        persist_outputs: PersistOutputsOptions | None = None,
        restore_outputs: RestoreOutputsOptions | None = None,
        result_key: str | None = None,
        fixed_params: Mapping[str, object] | None = None,
        param_key_map: Mapping[str, str] | None = None,
        **kwargs: object,
    ) -> Callable[[Callable], JobTemplateT]:
        ...


class IsTaskTemplateArgsJobTemplate(IsFuncArgJobTemplate[JobTemplateT, JobT],
                                    IsTaskTemplateArgsJobBase[TaskTemplateT],
                                    Protocol[TaskTemplateT, JobTemplateT, JobT]):
    """"""
    def refine(self,
               *task_templates: TaskTemplateT,
               update: bool = True,
               name: str | None = None,
               iterate_over_data_files: bool = False,
               output_dataset_param: str | None = None,
               output_dataset_cls: type[IsDataset] | None = None,
               auto_async: bool = True,
               fixed_params: Mapping[str, object] | None = None,
               param_key_map: Mapping[str, str] | None = None,
               result_key: str | None = None,
               persist_outputs: PersistOutputsOptions | None = None,
               restore_outputs: RestoreOutputsOptions | None = None,
               **kwargs: object) -> JobTemplateT:
        ...



from typing import Generic, get_args

from typing_extensions import TypeVar

from omnipy.data.helpers import TypeVarStore1, TypeVarStore2, TypeVarStore3, TypeVarStore4
from omnipy.data.model import Model
from omnipy.util.helpers import is_iterable


class NotIterableExceptStrOrBytesModel(Model[object | None]):
    """
    Model describing any object that is not iterable, except for `str` and `bytes` types.
    As strings and bytes are iterable (over the characters/bytes) but also generally useful and
    often considered singular (or scalar) types, they are specifically allowed by this model.

    Examples:
        >>> from omnipy import NotIterableExceptStrOrBytesModel, print_exception
        >>>
        >>> NotIterableExceptStrOrBytesModel(1234)
        NotIterableExceptStrOrBytesModel(1234)
        >>> NotIterableExceptStrOrBytesModel('1234')
        NotIterableExceptStrOrBytesModel(1234)
        >>> with print_exception:
        ...     NotIterableExceptStrOrBytesModel((1, 2, 3, 4))
        ValidationError: 1 validation error for NotIterableExceptStrOrBytesModel

    Note:
        JsonScalarModel is a strict submodel of NotIterableExceptStrOrBytesModel in that all objects
        allowed by JsonScalarModel are also allowed by NotIterableExceptStrOrBytesModel.
    """
    @classmethod
    def _parse_data(cls, data: object) -> object:
        if isinstance(data, NotIterableExceptStrOrBytesModel):
            return data

        assert isinstance(data, str) or isinstance(data, bytes) or not is_iterable(data), \
            f'Data of type {type(data)} is iterable'

        return data


#
# Exportable models
#

# General

U = TypeVar('U', bound=Model, default=Model[object])
V = TypeVar('V', bound=Model, default=Model[object])
W = TypeVar('W', bound=Model, default=Model[object])
X = TypeVar('X', bound=Model, default=Model[object])
Y = TypeVar('Y', bound=Model, default=Model[object])
Z = TypeVar('Z', bound=Model, default=Model[object])


class ChainMixin:
    @classmethod
    def _parse_data(cls, data: object) -> object:
        type_args = get_args(cls.outer_type(with_args=True))
        store_models = [get_args(store)[0] for store in type_args[1:-1]]
        all_models = [type_args[0]] + store_models + [type_args[-1]]
        all_models.reverse()

        if isinstance(data, all_models[-1]):
            return data

        assert isinstance(data, all_models[0]), \
            f'Expected data of type {all_models[0]}, got {type(data)}'

        for model in all_models[1:]:
            data = model(data)
        return data


class Chain2(ChainMixin, Model[V | U], Generic[U, V]):
    ...


class Chain3(ChainMixin, Model[W | TypeVarStore1[V] | U], Generic[U, V, W]):
    ...


class Chain4(ChainMixin, Model[X | TypeVarStore2[W] | TypeVarStore1[V] | U], Generic[U, V, W, X]):
    ...


class Chain5(
        ChainMixin,
        Model[Y | TypeVarStore3[X] | TypeVarStore2[W] | TypeVarStore1[V] | U],
        Generic[U, V, W, X, Y],
):
    ...


class Chain6(
        ChainMixin,
        Model[Z | TypeVarStore4[Y] | TypeVarStore3[X] | TypeVarStore2[W] | TypeVarStore1[V] | U],
        Generic[U, V, W, X, Y, Z],
):
    ...



from typing import Generic

from omnipy.data.dataset import Dataset
from omnipy.modules.frozen.models import (NestedFrozenDictsOrTuplesModel,
                                          NestedFrozenOnlyDictsModel,
                                          NestedFrozenOnlyTuplesModel)
from omnipy.modules.frozen.typedefs import KeyT, ValT


class NestedFrozenDictsOrTuplesDataset(Dataset[NestedFrozenDictsOrTuplesModel[KeyT, ValT]],
                                       Generic[KeyT, ValT]):
    ...


class NestedFrozenTuplesDataset(Dataset[NestedFrozenOnlyTuplesModel[ValT]], Generic[ValT]):
    ...


class NestedFrozenDictsDataset(Dataset[NestedFrozenOnlyDictsModel[KeyT, ValT]], Generic[KeyT,
                                                                                        ValT]):
    ...
